/* IBM Confidential
 * OCO Source Materials
 * 5724-C11, 5724-D17,5724-E30
 * (C) Copyright IBM Corporation 1997,2002
 * The source code for this program is not published or otherwise
 * divested of its trade secrets, irrespective of what has been
 * deposited with the U.S. Copyright Office.
*/


package com.crossworlds.DataHandlers.text;

import java.util.List;
import java.util.LinkedList;
import java.util.StringTokenizer;

import com.crossworlds.DataHandlers.DataHandler;
import com.crossworlds.DataHandlers.Exceptions.CW_BOFormatException;
import com.crossworlds.DataHandlers.Exceptions.CW_EndOfBOException;
import com.crossworlds.DataHandlers.Exceptions.CW_SubscriptionException;


//CW classes
import CxCommon.Exceptions.BusObjSpecNameNotFoundException;
import CxCommon.Exceptions.BusObjInvalidVerbException;
import CxCommon.Exceptions.CxObjectInvalidAttrException;
import CxCommon.Exceptions.CxObjectNoSuchAttributeException;
import CxCommon.BusinessObjectInterface;
import CxCommon.CxObjectContainerInterface;
import CxCommon.CxObjectAttr;
import AppSide_Connector.JavaConnectorUtil;

// java classes
//import java.io.*;
import java.io.InputStream;
import java.io.ByteArrayInputStream;
import java.io.Reader;

/**
*   Class which provides conversions between delimited data and a CrossWorlds
*   business object.  Any CrossWorlds business object can be used.  The class
*   expects that the business object name and verb appear as the first two
*   tokens for each object in the data.  Also, the data should contain a token
*   which gives the number of instances of child objects for each child which
*   appears in the business object.
*/
public class delimited extends DataHandler
{
    /**
    *   Used to build the delimited string for conversions to delimited format
    *   from a business object.
    */
    protected StringBuffer mBOStringBuffer = null;

    protected final int mBufferSize = 2048;  //2K Characters
    protected final int mAttrValueBufferSize = 256;
    protected final char TAB_CHAR = '\t';

    /**
     * CR12981
     * Used to read in a Stream into a char array when converting to
     * a business object.
     */
    protected char[] mBOCharBuffer;
    protected int mCurrentIndex;

    // Configuration properties.  These are either set based on options set
    // from config object(s) or defaulted.
    // CR9251 - Removed all of the transient keywords
    protected String mCxIgnore;
    protected String mCxBlank;
    protected String mDelimiter;
    protected String mEscape;

    // CR12981 the char delimiter and escape character are used
    // when these properties are single characters, not Strings.
    protected char mCharDelimiter;
    protected char mCharEscape;
    protected boolean isCharDelimiter = false;

    //for BOString parsing
    protected int mStartIndex;  //of a line
    protected int mTokenNum;     //for error messaging
    protected boolean mEndOfBOString;  //set in this.getNextLine()

    protected long totalTime = 0;

    public delimited()
    {
    }

    /**
    *   Sets the member variables according to the properties set by the base
    *   class based on the config object.
    */
    protected void getConfigOptions() throws Exception
    {
	traceWrite( "Entering getConfigOptions", JavaConnectorUtil.LEVEL4);

	mCxIgnore = getOption("CxIgnore");
	traceWrite( "CxIgnore value set to: " + mCxIgnore,
		JavaConnectorUtil.LEVEL5);

	mCxBlank = getOption("CxBlank");
	traceWrite( "CxBlank value set to: " + mCxBlank,
		    JavaConnectorUtil.LEVEL5);
	//CR 7563. To make tab and any other string as a valid delimiter the following code is modified by Madhavi.
	String lvDelimiter = getOption("Delimiter");

	if( lvDelimiter.length() > 0 )
	{
	    if ( lvDelimiter.equalsIgnoreCase("\\t") || lvDelimiter.equalsIgnoreCase("tab") )
	    {
		mCharDelimiter = TAB_CHAR;
		mDelimiter = "\t";
		isCharDelimiter = true;
	    }
	    else if ( lvDelimiter.length() == 1)
	    {
		mCharDelimiter = lvDelimiter.charAt(0);
		mDelimiter = lvDelimiter;
		isCharDelimiter = true;
	    }
	    else
	    {
		mDelimiter = lvDelimiter;
	    }
	}
	else
	{
	  mCharDelimiter = '~';
	  mDelimiter = "~";
	  isCharDelimiter = true;
	}

	traceWrite( "Delimiter value set to: " + mDelimiter, JavaConnectorUtil.LEVEL5);

	String lvEscapeCharacter = getOption("Escape");
	if( lvEscapeCharacter.length() > 0 )
	{
	  if (lvEscapeCharacter.length() == 1)
	  {
	    mCharEscape = lvEscapeCharacter.charAt(0);
	    mEscape = lvEscapeCharacter;
	  }
	  else
	  {
	    throw new Exception("The DataHandler MO Property, \"Escape\", must be a single character");
	  }
	}
	else
	{
	  mCharEscape = '\\';
	  mEscape = "\\";
	}

	traceWrite( "Escape character set to: " + mEscape, JavaConnectorUtil.LEVEL5);

	traceWrite( "Exiting getConfigOptions", JavaConnectorUtil.LEVEL4);
    }

    /**
    *   Resets member variables.  Should be called each time a method in the
    *   class is called to do a conversion to or from a business object.
    */
    protected void clear(Object config)
	throws Exception
    {
	traceWrite( "Entering clear() method", JavaConnectorUtil.LEVEL4);

	if ( config instanceof BusinessObjectInterface )
	{
	    traceWrite( "Setting options based on config object attributes",
			JavaConnectorUtil.LEVEL4);
	    setupOptions( (BusinessObjectInterface)config);
	}

	getConfigOptions();

	mBOStringBuffer = new StringBuffer(100);
	mStartIndex = 0;
	mEndOfBOString = false;

	// CR12981
	mCurrentIndex = 0;

	traceWrite( "Exiting clear() method", JavaConnectorUtil.LEVEL4);
    }

    /**
    *   Converts the data contained in a Reader object to a CrossWorlds
    *   business object.  Allows configuration options to be set by passing
    *   in a meta object which contains the config information.
    *
    *   @param  serializedData  Reader object which contains the data to be
    *                           converted to a business object.
    *   @param  config          If this parameter is an instance of
    *                           BusinessObjectInterface config options will
    *                           be set based on the default values of the
    *                           attributes.
    */
    public BusinessObjectInterface getBO(Reader serializedData, Object config)
	throws Exception
    {
	traceWrite( "Entering getBO(Reader, Object)", JavaConnectorUtil.LEVEL4);

	clear(config);

	BusinessObjectInterface resultObj = null;

	// Create a String object from the Reader, then use the string based
	//  method
	int conversionCheck;
	char[] temp = new char[2000];
	StringBuffer tempStringBuffer = new StringBuffer(1000);

	while ( (conversionCheck = serializedData.read (temp)) != -1)
	  tempStringBuffer.append(temp, 0, conversionCheck);

	mBOCharBuffer = new char[tempStringBuffer.length()];
	tempStringBuffer.getChars(0, mBOCharBuffer.length, mBOCharBuffer, 0);

	if (JavaConnectorUtil.isTraceEnabled(JavaConnectorUtil.LEVEL5))
	{
	  String toWrite = new String(mBOCharBuffer);

	  traceWrite( "getBO(Reader, Object) set mBOCharBuffer to the following: \n"
		    + toWrite, JavaConnectorUtil.LEVEL5);
	}

	traceWrite("getBO(Reader, Object) calling getBOFromString",
		    JavaConnectorUtil.LEVEL5);
	resultObj = getBOFromString(null);
	// <I18N> by Chen Gang (07/22/2002)
	if(resultObj != null && getLocale() != null)
		resultObj.setLocale(getLocale());
	// </I18N>
	traceWrite( "Exiting getBO(Reader, Object)", JavaConnectorUtil.LEVEL4);
	// There is a performance issue with this because the dump is created
	//   and concatenated BEFORE the trace level can be checked
	// traceWrite( "Return value: \n" + resultObj.dump,
	//              JavaConnectorUtil.LEVEL5);
	return resultObj;

    }

    /**
    *   Populates a given BusinessObjectInterface with data extracted
    *   from Reader object. In this case it is up to the caller to create
    *   the BO.  Allows configuration options to be set by passing in a
    *   meta object which contains the config information.
    *
    *   @param  serializedData  Reader object which contains the data to
    *                           be converted to a business object.
    *   @param  theObj          BusinessObjectInterface to be populated
    *                           with data from Reader object
    *                           'serializedData'.
    *   @param  config          If this parameter is an instance of
    *                           BusinessObjectInterface config options will
    *                           be set based on the default values of the
    *                           attributes.
    */
    public void getBO(Reader serializedData, BusinessObjectInterface theObj,
		      Object config)
	throws Exception
    {
	traceWrite( "Entering getBO(Reader, BusinessObjectInterface, Object)",
		    JavaConnectorUtil.LEVEL4);

	clear(config);

	int conversionCheck;
	char[] temp = new char[2000];
	StringBuffer tempStringBuffer = new StringBuffer(1000);

	while ( (conversionCheck = serializedData.read (temp)) != -1)
	  tempStringBuffer.append(temp, 0, conversionCheck);

	mBOCharBuffer = new char[tempStringBuffer.length()];
	tempStringBuffer.getChars(0, mBOCharBuffer.length, mBOCharBuffer, 0);

	traceWrite( "getBO(Reader, BusinessObjectInterface, Object) set mBOCharBuffer to the following: \n"
		    + mBOCharBuffer, JavaConnectorUtil.LEVEL5);

	traceWrite("getBO(Reader, BusinessObjectInterface, Object) calling getBOFromString",
		    JavaConnectorUtil.LEVEL5);

	// Pass BO by reference
	getBOFromString( theObj );
	// <I18N> by Chen Gang (07/22/2002)
	if(theObj != null && getLocale() != null)
	  theObj.setLocale(getLocale());
	// </I18N>
	traceWrite( "Exiting getBO(Reader, BusinessObjectInterface, Object)",
		    JavaConnectorUtil.LEVEL4);
	// There is a performance issue with this because the dump is created
	//   and concatenated BEFORE the trace level can be checked
	// traceWrite( "Return value: \n" + theObj.dump(),
	//              JavaConnectorUtil.LEVEL5);

    }

    /**
    *   Populates a given BusinessObjectInterface with data extracted
    *   from String object. In this case it is up to the caller to create
    *   the BO.  Allows configuration options to be set by passing in a meta
    *   object which contains the config information.
    *
    *   @param  serializedData  String object which contains the data to be
    *                           converted to a business object.
    *   @param  theObj          BusinessObjectInterface to be populated with
    *                           data from Reader object 'serializedData'.
    *   @param  config          If this parameter is an instance of
    *                           BusinessObjectInterface config options will be
    *                           set based on the default values of the
    *                           attributes.
    */
    public void getBO(String serializedData, BusinessObjectInterface theObj,
			Object config )
	throws Exception
    {
	traceWrite( "Entering getBO(String, BusinessObjectInterface, Object)",
		    JavaConnectorUtil.LEVEL4);

	clear(config);

	try
	{
	  mBOCharBuffer = new char[serializedData.length()];
	  serializedData.getChars(0, mBOCharBuffer.length, mBOCharBuffer, 0);

	  traceWrite("getBO(String, BusinessObjectInterface, Object) calling getBOFromString",
			JavaConnectorUtil.LEVEL5);

	  // Pass BO by reference
	  getBOFromString( theObj );
	}
	catch(Exception e)
	{
	    traceWrite( "Error occured in getBO method: " + e.toString(), 1);
	    throw e;
	}
	// <I18N> by Chen Gang (07/22/2002)
	if(theObj != null && getLocale() != null)
	  theObj.setLocale(getLocale());
	// </I18N>


	traceWrite( "Exiting getBO(String, BusinessObjectInterface, Object)",
		    JavaConnectorUtil.LEVEL4);
	// There is a performance issue with this because the dump is created
	//   and concatenated BEFORE the trace level can be checked
	// traceWrite( "Return value: \n" + theObj.dump(),
	//              JavaConnectorUtil.LEVEL5);

    }

    /**
    *   Populates a given BusinessObjectInterface with data extracted
    *   from String object. In this case it is up to the caller to create
    *   the BO.  Allows configuration options to be set by passing in a meta
    *   object which contains the config information.
    *
    *   @param  serializedData  String object which contains the data to be
    *                           converted to a business object.
    *   @param  config          If this parameter is an instance of
    *                           BusinessObjectInterface config options will be
    *                           set based on the default values of the
    *                           attributes.
    */
    public BusinessObjectInterface getBO(String serializedData, Object config )
	throws Exception
    {
	traceWrite( "Entering getBO(String, Object)",
		    JavaConnectorUtil.LEVEL4);
	clear(config);

	BusinessObjectInterface resultObj = null;

	try
	{
	  mBOCharBuffer = new char[serializedData.length()];
	  serializedData.getChars(0, mBOCharBuffer.length, mBOCharBuffer, 0);

	  traceWrite("getBO(String, BusinessObjectInterface, Object) calling getBOFromString",
			JavaConnectorUtil.LEVEL5);

	  // Pass BO by reference
	  resultObj = getBOFromString(null );
	}
	catch(Exception e)
	{
	    traceWrite( "Error occured in getBO method: " + e.toString(), 1);
	    throw e;
	}
	// <I18N> by Chen Gang (07/22/2002)
	if(resultObj != null && getLocale() != null)
	  resultObj.setLocale(getLocale());
	// </I18N>

	traceWrite( "Exiting getBO(String, BusinessObjectInterface, Object)",
		    JavaConnectorUtil.LEVEL4);

	return resultObj;



    }

    /**
    *   Creates an instance of BusinessObjectInterface and populates the verb.
    *   Calls parseAttributeList to do the parsing of the String and populate
    *   the attribute values.  Determines what type of business object to
    *   create either based on the paramter value or on the first token in the
    *   data.  The paramter value must match the first token in the data,
    *   otherwise an error is generated.
    *
    *   @param  pvBOType    This is the type of Business Object that should be
    *                       created.  If this is not null it is checked against
    *                       the first token in the data, and an error is
    *                       generated if there is a mis-match.
    */
    private BusinessObjectInterface getBOFromString(BusinessObjectInterface pvBO)
	throws Exception
    {
	traceWrite("Entering getBOFromString(String)",
		    JavaConnectorUtil.LEVEL4);

	// BO has been passed by reference, no need to create new BO
	/* BusinessObjectInterface lvResultBO = null; */

	String pvBOType = null;
	if (pvBO != null) {
	  pvBOType = pvBO.getName();
	}

	String lvBOName = null;
	String lvVerb = null;

	try
	{

	    // Get BO name and verb - assumes they are the first two tokens
	    //  in the delmited string
	    lvBOName = getNextToken();
	    traceWrite( "BusinessObject name from the message contents: "
			+ lvBOName,
			JavaConnectorUtil.LEVEL5);
	    lvVerb = getNextToken();
	    traceWrite( "BusinessObject verb from the message contnents: "
			+ lvVerb,
			JavaConnectorUtil.LEVEL5);

	    //check of given to the BO name
	    if (pvBOType != null && ! pvBOType.equals(lvBOName)) {
		throw new CW_BOFormatException(" Type miss-match at: column = "
			+ (getStartIndex(lvVerb) - lvBOName.length())
			+ ": BO type miss-match is the expected type; expects "
			+ pvBOType);
	    }
	    // Create new BO is passed a null pointer as pvBO
	    else if (pvBO == null) {
		pvBO = JavaConnectorUtil.createBusinessObject (lvBOName);
	    }

	    //create new BO and continue the parse the rest
	    //lvResultBO = JavaConnectorUtil.createBusinessObject(lvBOName);
	    if (! mCxIgnore.equals( lvVerb)) {
		pvBO.setVerb(lvVerb);
	    }

	    //now parse through the main BO's attributes
	    parseAttributeList(pvBO);
	}
	catch (BusObjSpecNameNotFoundException e)
	{
	    throw new CW_BOFormatException(" at column = "
		+ this.getStartIndex(lvBOName) + ": BO name of "
		+ lvBOName + " not found: " +  e.toString());
	}

	catch (BusObjInvalidVerbException e)
	{
	  throw new CW_BOFormatException(" at column = "
	    + this.getStartIndex(lvVerb) + ": Verb " + lvVerb
	    + " of BO name " + lvBOName + " not found: "
	    + e.toString());
	}

	traceWrite( "Exiting getBOFromString(String)",
		    JavaConnectorUtil.LEVEL4);
	return pvBO;

    } // end getBOFromString

    /**
    *   Creates a String based on the data contained in a BusinessObjectInterface.
    *   Any valid CrossWorlds business object can be used.
    *   Allows configuration options to be set by passing in a meta object
    *   which contains the config information.
    *
    *   @param  theObj      BusinessObjectInterface which contains the data to
    *                       be included in the String.
    *   @param  config      If an instance of BusinessObjectInterface is
    *                       passed in, the default values of the attributes
    *                       are used as options for the formatter.
    *   @return     String which contains the data from theObj in delimited
    *               format.
    */
    public String getStringFromBO(BusinessObjectInterface theObj,
				  Object config)
      throws Exception
    {
	traceWrite( "Entering getStringFromBO(BusinessObjectInterface, Object)",
		    JavaConnectorUtil.LEVEL4);

	clear(config);

	BusinessObjectInterface returnObj = null;

	traceWrite( "Calling setAttrList(BusinessObjectInterface) from getStringFromBO",
			JavaConnectorUtil.LEVEL5);
	this.setAttrList(theObj);

	traceWrite( "Exiting getStringFromBO(BusinessObjectInterface, Object)",
		    JavaConnectorUtil.LEVEL4);
	return mBOStringBuffer.toString();
    } // end getStringFromBO

    /**
    *   Creates an InputStream based on the data contained in a BusinessObjectInterface.
    *   Creates a String object first, then converts it to an InputStream to
    *   return to the caller.
    *   Any valid CrossWorlds business object can be used.
    *   Allows configuration options to be set by passing in a meta object
    *   which contains the config information.
    *
    *   @param  theObj      BusinessObjectInterface which contains the data to
    *                       be included in the String.
    *   @param  config      If an instance of BusinessObjectInterface is
    *                       passed in, the default values of the attributes
    *                       are used as options for the formatter.
    *   @return     InputStream which contains the data from theObj in
    *               delimited format.
    */
    public InputStream getStreamFromBO(BusinessObjectInterface theObj,
					Object config)
	throws Exception
    {
	traceWrite( "Entering getStreamFromBO(BusinessObjectInterface, Object)",
		    JavaConnectorUtil.LEVEL4);

	String BOstring;

	traceWrite( "getStreamFromBO calling getStringFromBO",
		    JavaConnectorUtil.LEVEL5);
	BOstring = getStringFromBO(theObj, config);

	traceWrite( "Exiting getStreamFromBO(BusinessObjectInterface, Object)",
		    JavaConnectorUtil.LEVEL4);

	// <I18N> by Chen Gang (07/22/2002)
	// return new ByteArrayInputStream( BOstring.getBytes() );
	if(getEncoding() != null) {
	  return new ByteArrayInputStream( BOstring.getBytes(getEncoding()));
	} else {
	  return new ByteArrayInputStream( BOstring.getBytes());
	}
	// </I18N>
    } // end getStreamFromBO

    /**
    *   Adds a token value and the proper delimiter to the data String
    *   being created (mBOStringBuffer).  Also contains the logic to check
    *   for the escape and delimiter character in the token value passed in
    *   if needed.
    *
    *   @param  pvCheckDelimiter    Determines whether or not logic to check
    *                               token value for delimiter and escape
    *                               character is performed.  If value is 'true',
    *                               logic is performed.  If value is 'false'
    *                               logic is not performed.
    *   @param  pvTokenValue        Token value to be added to the string
    *                               being built.
    */
    protected void setSimpleToken(boolean pvCheckDelimiter,
				  String pvTokenValue)
    {
	traceWrite( "Entering setSimpleToken(boolean, String)",
		    JavaConnectorUtil.LEVEL4);
	traceWrite( "pvCheckDelimiter = " + pvCheckDelimiter
		    + "\n pvTokenValue = " + pvTokenValue,
		    JavaConnectorUtil.LEVEL5);

	if (pvTokenValue == null)
	    mBOStringBuffer.append(mCxIgnore);
	else if (pvTokenValue.equals(""))
	    mBOStringBuffer.append(mCxBlank);
	else
	{
	    if (pvCheckDelimiter)
	    {
		int lvDelimiterIndex = 0;
		int lvEscapeIndex = 0;
		int lvSearchHeadIndex = 0;
		int lvTokenLength = pvTokenValue.length();

		do
		{
		    lvDelimiterIndex = pvTokenValue.indexOf(mDelimiter,
							    lvSearchHeadIndex);
		    lvEscapeIndex =  pvTokenValue.indexOf(mEscape,
							  lvSearchHeadIndex);
		    if (lvEscapeIndex == -1) //escape charactor is not present in pvTokenValue
		    {
			if (lvDelimiterIndex > -1) //delimiter found
			{
			    mBOStringBuffer.append(pvTokenValue.substring(lvSearchHeadIndex, lvDelimiterIndex));
			    mBOStringBuffer.append(mEscape);
			    mBOStringBuffer.append(mDelimiter);
			    lvSearchHeadIndex = lvDelimiterIndex + mDelimiter.length();
			    //lvSearchHeadIndex = lvDelimiterIndex + 1;
			}
			else  //both delimiter and escape are not found
			{
			    mBOStringBuffer.append(pvTokenValue.substring(lvSearchHeadIndex,lvTokenLength));
			    lvSearchHeadIndex = lvTokenLength + 1;
			}
		    }
		    else //escape is found
		    {
			if (lvDelimiterIndex > -1 )  //delimiter found
			{
			    if( lvEscapeIndex >= lvDelimiterIndex)
			    {
				//append the upto the one with smaller index first
				mBOStringBuffer.append(pvTokenValue.substring(lvSearchHeadIndex, lvDelimiterIndex));
				mBOStringBuffer.append(mEscape);
				mBOStringBuffer.append(mDelimiter);
				lvSearchHeadIndex = lvDelimiterIndex + mDelimiter.length();
				//lvSearchHeadIndex = lvDelimiterIndex + 1;
			    }
			    else
			    {
				mBOStringBuffer.append(pvTokenValue.substring(lvSearchHeadIndex, lvEscapeIndex));
				mBOStringBuffer.append(mEscape);
				mBOStringBuffer.append(mEscape);
				lvSearchHeadIndex = lvEscapeIndex + mEscape.length();
				//lvSearchHeadIndex = lvEscapeIndex + 1;
			    }
			}
			else  // if (lvDelimiterIndex > -1 )  //delimiter NOT found
			//only escape is found
			{
			    mBOStringBuffer.append(pvTokenValue.substring(lvSearchHeadIndex, lvEscapeIndex));
			    mBOStringBuffer.append(mEscape);
			    mBOStringBuffer.append(mEscape);
			    lvSearchHeadIndex = lvEscapeIndex + mEscape.length();
			    //lvSearchHeadIndex = lvEscapeIndex + 1;
			}
		    }
		}
		while(lvSearchHeadIndex < lvTokenLength);  //not in the token value
	    }
	    else
		mBOStringBuffer.append(pvTokenValue);
	}
	mBOStringBuffer.append(mDelimiter);

	traceWrite( "Exiting setSimpleToken", JavaConnectorUtil.LEVEL4);
    } // end setSimpleToken

    /**
    *   Loops through the attributes in a BusinssObjectInterface and add the
    *   attribute values to the mBOStringBuffer.
    *
    *   @param  pvBO    BusinessObjectInterface which contains the data to
    *                   be added to the delimited String.
    */
    protected void setAttrList(BusinessObjectInterface pvBO)
	throws CW_BOFormatException
    {
	traceWrite( "Entering setAttrList for business object: "
		    + pvBO.getName(),
		    JavaConnectorUtil.LEVEL4);

	String lvBOName = null;
	String lvBOVerb = null;

	lvBOName = pvBO.getName();
	lvBOVerb = pvBO.getVerb();

	// The business object name and verb are added as the first two
	//  tokens in the delimited string.
	this.setSimpleToken(true ,lvBOName);
	this.setSimpleToken(true ,lvBOVerb);

	List moAttrNames = listMOAttrNames(pvBO);
	String lvAttrName = null;
	try
	{
	    int lvAttrCount = 0;
	    if ( getOption("OmitObjectEventId").equalsIgnoreCase("true") )
	    {
	      lvAttrCount = pvBO.getAttrCount()-1; //skip objectEventId
	    }
	    else
	    {
	      lvAttrCount = pvBO.getAttrCount();
	    }

	    // Loop through the attributes in the business object to add
	    //  the value of each to the delimited string.
	    ATTRIBUTE_WALK: for (int lvAttrIndex = 0; lvAttrIndex < lvAttrCount; ++lvAttrIndex)
	    {
		CxObjectAttr lvAttrSpec = pvBO.getAttrDesc(lvAttrIndex);
		lvAttrName = lvAttrSpec.getName();

		// Check if the attribute is a simple attribute, or a
		//  contained BO
		if (lvAttrSpec.isObjectType())
		{
		    //skip child objects designated as meta objects
		    if( moAttrNames.contains( lvAttrName ) ) {
		      continue ATTRIBUTE_WALK;
		    }

		    // Check the cardinality of the container
		    if (lvAttrSpec.isMultipleCard())
		    {
			// Container is multiple cardinality
			CxObjectContainerInterface lvAttrMultiCardBOValue =
			    (CxObjectContainerInterface) pvBO.getAttrValue(lvAttrIndex);

			if (lvAttrMultiCardBOValue == null)
			{
			    traceWrite( "setAttrList found empty multiple cardinality container " + lvAttrSpec.getTypeName(),
					JavaConnectorUtil.LEVEL5);
			    // Add the object count to the delimited string
			    this.setSimpleToken(false, "0");
			}
			else
			{
			    int lvObjectCount = lvAttrMultiCardBOValue.getObjectCount();
			    traceWrite( "setAttrList found multiple cardinality container " + lvAttrSpec.getTypeName()
					+ " with " + lvObjectCount + " instances",
					JavaConnectorUtil.LEVEL5);

			    // Add the object count to the delimited string
			    this.setSimpleToken(false, Integer.toString(lvObjectCount));
			    // Loop through the object container, and make the recursive call for each
			    //  to add each object to the delimited string
			    for (int lvContObjectIndex = 0; lvContObjectIndex < lvObjectCount; ++lvContObjectIndex)
				setAttrList(lvAttrMultiCardBOValue.getBusinessObject(lvContObjectIndex));
			}
		    }
		    else
		    {
			// Container is single cardinality
			BusinessObjectInterface lvAttrSingleCardBOValue = (BusinessObjectInterface) pvBO.getAttrValue(lvAttrIndex);

			if (lvAttrSingleCardBOValue == null)
			{
			    traceWrite( "setAttrList found empty single cardinality container " + lvAttrSpec.getTypeName(),
					JavaConnectorUtil.LEVEL5);
			    // Add the object count to the delimited string
			    this.setSimpleToken(false, "0"); //object count number
			}
			else
			{
			    traceWrite( "setAttrList found single cardinality container " + lvAttrSpec.getTypeName(),
					JavaConnectorUtil.LEVEL5);
			    // Add the object count to the delmited string
			    this.setSimpleToken(false, "1");
			    // Make the recursive call to add the values from the contained object to the
			    //  delimited string
			    setAttrList(lvAttrSingleCardBOValue);
			}
		    }
		}
		else
		    this.setSimpleToken(true, (String)pvBO.getAttrValue(lvAttrIndex));
	    }
	}
	catch (CxObjectNoSuchAttributeException e)
	{
	    throw new CW_BOFormatException(e.toString());
	}

	traceWrite( "Exiting setAttrList for business object " + pvBO.getName(),
		    JavaConnectorUtil.LEVEL4);
    } // end setAttrList

    /**
    *   Parses the BO String and matches the tokens with attributes in the
    *   business object.
    *
    *   @param  pvBO    BusinessObjectInterface which will be populated with
    *                   the data from the delimited business object string.
    */
    protected void parseAttributeList(BusinessObjectInterface pvBO)
	throws CW_BOFormatException, CW_EndOfBOException, CW_SubscriptionException
    {
	traceWrite( "Entering parseAttributeList for business object "
		    + pvBO.getName(),
		    JavaConnectorUtil.LEVEL4);

	// Assumption is made that if this method is called, there should
	//  be some part of the string left to parse, otherwise we generate
	//  an error.
	if (mEndOfBOString || pvBO == null)
	    throw new CW_EndOfBOException(" Column: " + mStartIndex
		+ ":  immature ending of BO.  Please check your data format; ");


	int lvAttrNum = 0;
	if ( getOption("OmitObjectEventId").equalsIgnoreCase("true") )
	  lvAttrNum = pvBO.getAttrCount()-1; // skip Object event id
	else
	  lvAttrNum = pvBO.getAttrCount();

	String lvAttrValue = null;
	// Used only for the error message in case of an exception
	String lvAttrName = null;

	try
	{
	    List moAttrNames = listMOAttrNames( pvBO );

	    ATTRIBUTE_WALK: for (int lvAttrIndex = 0; lvAttrIndex < lvAttrNum; lvAttrIndex++)
	    {
		CxObjectAttr lvAttrSpec = pvBO.getAttrDesc(lvAttrIndex);
		lvAttrName = lvAttrSpec.getName();

		//skip child objects designated as meta objects
		if( moAttrNames.contains( lvAttrName ) ) {
		  continue ATTRIBUTE_WALK;
		}

		lvAttrValue = getNextToken();

		// Check to see if the attribute is a simple attribute,
		//  or is a business object container.
		if (lvAttrSpec.isObjectType())
		{
		    String lvBOType = lvAttrSpec.getTypeName();

		    if (lvAttrSpec.isMultipleCard())
		    {
			traceWrite( "parseAttributeList found multiple cardinality container for "
				    + lvBOType,
				    JavaConnectorUtil.LEVEL5);
			this.getMultipleCard(pvBO, lvAttrIndex, lvBOType, lvAttrValue);
		    }
		    else
		    {
			traceWrite( "parseAttributeList found single cardinality container for "
				    + lvBOType,
				    JavaConnectorUtil.LEVEL5);
			this.getSingleCard(pvBO, lvAttrIndex, lvBOType, lvAttrValue);
		    }
		}
		else
		{
		    // For simple (String) attribute, we check to see if the token is either
		    //  null, or matches the configured CxIgnore or CxBlank values.
		    //  Otherwise, set the attribute value with the token value.
		    if (lvAttrValue == null || lvAttrValue.equals(mCxIgnore) || lvAttrValue.equals(""))
			pvBO.setAttrValue(lvAttrIndex, null);   //this includes ObjectEventIdAttr
		    else if (lvAttrValue.equals(mCxBlank) || lvAttrValue.equals(" "))
			pvBO.setAttrValue(lvAttrIndex, "");   //this includes ObjectEventIdAttr
		    else
			pvBO.setAttrValue(lvAttrIndex, lvAttrValue);   //this includes ObjectEventIdAttr
		}
	    }
	}
	catch (CxObjectNoSuchAttributeException e)
	{
	    throw new CW_BOFormatException(e.toString() + ": data format error; Attribute miss alignment at: attribute name = " +
		lvAttrName + "; column = " + getStartIndex(lvAttrValue) );
	}
	catch (CW_SubscriptionException e)
	{
	    throw new CW_BOFormatException(e.toString() + ": subscription error; Attribute miss alignment at: attribute name = " +
		lvAttrName + "; column = " + getStartIndex(lvAttrValue) );
	}
	catch (CxObjectInvalidAttrException e)
	{
	    throw new CW_BOFormatException(e.toString() + ": data format error; Attribute miss alignment at: attribute name = " +
		lvAttrName + "; column = " + getStartIndex(lvAttrValue) );
	}

	traceWrite( "Exiting parseAttributeList for business object " + pvBO.getName(),
		    JavaConnectorUtil.LEVEL4);

    } // end parseAttributeList

    /**
    *   Populates a single cardinatlity child business object with the token
    *   values found in the delimited string.
    *
    *   @param  pvParentBO              Parent business object.
    *   @param  pvParentAttrIndex       Attribute index of the child to be
    *                                   created.
    *   @param  pvBOType                Type of the child business object.
    *   @param  pvObjectCountString     Number of instances of the child
    *                                   object as parsed from the delimited
    *                                   string.
    */
    protected void  getSingleCard(BusinessObjectInterface pvParentBO,
				  int pvParentAttrIndex, String pvBOType,
				  String pvObjectCountString)
	throws CW_BOFormatException, CW_EndOfBOException,
		CW_SubscriptionException
    {
	traceWrite( "Entering getSingleCard for parentBO "
		    + pvParentBO.getName() + " index " + pvParentAttrIndex
		    + " with count " + pvObjectCountString,
		    JavaConnectorUtil.LEVEL4);

	String childBOName = null;
	String childVerb = null;

	try
	{
	    BusinessObjectInterface lvChildBO = null;
	    // Check the object count token
	    // If it does not match "1", assume that the child object should
	    //  be null
	    if (pvObjectCountString.equals("1"))
	    {
		// The delimited string contains a single instance of the
		//  child object
		traceWrite( "Found single instance of child object "
			    + pvBOType + " in delimited string",
			    JavaConnectorUtil.LEVEL5);

		childBOName = getNextToken();
		if (! pvBOType.equals(childBOName) )
		{
		    throw new CW_BOFormatException(" Type miss-match at: column = "
			+ getStartIndex(childBOName)
			+ ": BO type miss-match is the expected type; expects "
			+ pvBOType);
		}
		else
		{
		    lvChildBO = JavaConnectorUtil.createBusinessObject(pvBOType);
		    childVerb = getNextToken();

		    if(! mCxIgnore.equals(childVerb) && ! mCxBlank.equals(childVerb))
			lvChildBO.setVerb(childVerb);

		    parseAttributeList(lvChildBO);
		    pvParentBO.setAttrValue(pvParentAttrIndex, lvChildBO);
		}
	    }
	    else if ( pvObjectCountString.equals(mCxIgnore) ||
			pvObjectCountString.equals("0"))
	    {
		// This case is included to validate the object count token is valid
		traceWrite( "Encountered null child object " + pvBOType
			    + " for parent " + pvParentBO.getName(),
			    JavaConnectorUtil.LEVEL5);
	    }
	    else
		throw new CW_BOFormatException(" Column: "
			+ getStartIndex(pvObjectCountString) +
			" BO object count error.  This count should be 1, 0, or CxIgnore. " );

	}
	catch (CxObjectInvalidAttrException e)
	{
	    throw new CW_BOFormatException(" Column: " +
					    getStartIndex(pvObjectCountString) + ": try to access to the wrong attribute. " +  e.getMessage() + ". ");
	}
	catch (CxObjectNoSuchAttributeException e)
	{
	    throw new CW_BOFormatException(" Column: " + getStartIndex(pvObjectCountString) +
					    ": try to access to the wrong attribute. " +  e.getMessage() + ". ");
	}
	catch (BusObjSpecNameNotFoundException e)
	{
	    throw new CW_BOFormatException("Invalid BusObjSpec name found: "
				+ e.getMessage());
	}
	catch (BusObjInvalidVerbException e)
	{
	    throw new CW_BOFormatException(" at column = "
		+ this.getStartIndex(childVerb) + ": Verb " + childVerb
		+ " of BO name " + childBOName + " not found: "
		+ e.toString());
	}

	traceWrite( "Exiting getSingleCard for parentBO " + pvParentBO.getName() + " index " + pvParentAttrIndex
		    + " with count " + pvObjectCountString,
		    JavaConnectorUtil.LEVEL4);
    } // end getSingleCard


    /**
    *   Populates a child container based on token values in the delimited
    *   string.
    *   Expects all objects in the child container to be of the same type.
    *
    *   @param  pvParentBO              Parent business object.
    *   @param  pvParentAttrIndex       Attribute index of the child to be
    *                                   created.
    *   @param  pvBOType                Type of the child business object.
    *   @param  pvObjectCountString     Number of instances of the child
    *                                   object as parsed from the delimited
    *                                   string.
    */
    protected void getMultipleCard(BusinessObjectInterface pvParentBO,
				   int pvParentAttrIndex, String pvBOType,
				   String pvObjectCountString)
	throws CW_BOFormatException, CW_EndOfBOException,
		CW_SubscriptionException
    {
	traceWrite( "Entering getMultipleCard for parentBO "
		    + pvParentBO.getName() + " index " + pvParentAttrIndex
		    + " with count " + pvObjectCountString,
		    JavaConnectorUtil.LEVEL4);

	String childBOName = null;
	String childVerb = null;

	try
	{
	    if ( pvObjectCountString.equals(mCxIgnore) )
	    {
		traceWrite( "Count string for container " + pvBOType
			    + " in parent " + pvParentBO.getName()
			    + " is CxIgnore value",
			    JavaConnectorUtil.LEVEL5);
	    }
	    else
	    {
		int lvObjectCount = Integer.parseInt(pvObjectCountString);

		if ( lvObjectCount == 0)
		{
		    traceWrite( "Count string for container " + pvBOType
				+ " in parent " + pvParentBO.getName()
				+ " is 0",
				JavaConnectorUtil.LEVEL5);
		}
		else if (lvObjectCount > 0)
		{
		    // There is at least one instance of the object in the
		    //  delimited string

		    traceWrite( "Found " + lvObjectCount
				+ " instances of child object pvBOType "
				+ " in parent object " + pvParentBO.getName(),
				JavaConnectorUtil.LEVEL5);

		    BusinessObjectInterface lvChildBO = null;

		    // For each instance of the child object, parse the
		    //  attribute list, and then add the object container
		    //  to the parent.
		    for (int lvObjectIndex = 0; lvObjectIndex < lvObjectCount; lvObjectIndex++)
		    {
			// CR12793 - The first call to getNextToken will remove the BOName.
			// It exists in the for loop because the BOName is added to every child
			// instance. The second call to getNextToken is needed so that the BOVerb
			// can be set appropriately. The verb is also added to every child instance.
			childBOName = getNextToken();
			childVerb = getNextToken();

			lvChildBO = JavaConnectorUtil.createBusinessObject(pvParentBO.getAttrDesc(pvParentAttrIndex).getTypeName());
			if(! mCxIgnore.equals(childVerb) && ! mCxBlank.equals(childVerb))
			  lvChildBO.setVerb(childVerb);

			parseAttributeList(lvChildBO);
			pvParentBO.setAttrValue(pvParentAttrIndex, lvChildBO);
		    }
		}
	    }
	}
	catch (CxObjectNoSuchAttributeException e)
	{
	    throw new CW_BOFormatException(" Column: " + getStartIndex(pvObjectCountString) +
					": formatting error; " + e.getMessage() + ". ");
	}
	catch (CxObjectInvalidAttrException e)
	{
	    throw new CW_BOFormatException(" Column: " + getStartIndex(pvObjectCountString) +
		": formatting error: Attribute miss alignment; " + e.getMessage() + ". ");
	}
	catch (NumberFormatException e)
	{
	    throw new CW_BOFormatException(" Column: " + getStartIndex(pvObjectCountString) +
		": formatting error:  cannot recognize number format; " + e.getMessage() + ". ");
	}
	catch (BusObjSpecNameNotFoundException e)
	{
	    throw new CW_BOFormatException("Invalid BusObjSpec name found: " + e.getMessage());
	}
	catch (BusObjInvalidVerbException e)
	{
	    throw new CW_BOFormatException(" at column = "
		+ this.getStartIndex(childVerb) + ": Verb " + childVerb
		+ " of BO name " + childBOName + " not found: "
		+ e.toString());
	}

	traceWrite( "Exiting getMultipleCard for parentBO " + pvParentBO.getName() + " index " + pvParentAttrIndex
		    + " with count " + pvObjectCountString,
		    JavaConnectorUtil.LEVEL4);

    } // end getMultipleCard

    /**
    *   Computes start index of current token
    *
    *   @param  pvTokenValue     String value of the current token
    *   @return     Starting index of the current token calculated by
    *               subtracting the current token size from the start index
    *               of the next token.
    */
    protected int getStartIndex(String pvTokenValue)
    {
	return (mStartIndex - pvTokenValue.length());
    }

    /**
    *   Parses the next token from mBOString based on the next appearance of
    *   the configured delimiter.  Also deletes the escape character from the
    *   token value where it immediately preceeds the delimiter.
    *
    *   @return         String value that is the next token that appers in the
    *                   mBOString.
    */
      protected String getNextToken() throws CW_EndOfBOException
      {
	  if (mCurrentIndex >= mBOCharBuffer.length)
	  {
	    throw new CW_EndOfBOException(" Already reached end of BO while expecting more token ");
	  }

	  String lvResult = null;

	  // CR12981 Check to see if the Delimiter can be
	  // represented as a single character. If so, follow
	  // a different code path for efficiency purposes.
	  // charDelimiterRead = efficient because parsing char
	  // stringDelmiterRead = less efficient because parsing string
	  if (isCharDelimiter)
	  {
	    lvResult = charDelimiterRead();
	  }
	  else
	  {
	    lvResult = stringDelimiterRead();
	  }
	  return lvResult;
      } // end getNextToken

    /**
     *  Performs search for char delimiter
     *  and returns string before the delimter.
     *  Uses literal char comparisons for searching.
     *
     *  @return         String value that is the next
     *                  token that appers in the
     *                  mBOCharBuffer.
     *
     *  Created for CR: CR12981
     */
    private String charDelimiterRead()
    {

      StringBuffer buffer = new StringBuffer(256);

	int startIndex = mCurrentIndex;
	int endIndex = startIndex;
	while ( endIndex < mBOCharBuffer.length ) {

	    // Check if this character is a delimiter
	    if ( mBOCharBuffer[endIndex] == mCharDelimiter) {
		// check previous character to see if it is an escape character
		if ( mBOCharBuffer[endIndex - 1] != mCharEscape )
		    // it isn't so we found an unescaped delimiter and thus the
		    // end of our token
		    break;
		else {
		    // it is an escaped delimiter, so save the portion up to
		    // the escaped character in our buffer
		    buffer.append(mBOCharBuffer, startIndex, endIndex  - 1 -
			mCurrentIndex);

		    // append the delimiter in the buffer ( since it was an
		    // escaped delimiter
		    buffer.append(mCharDelimiter);

		    // move the startIndex to the character past the delimiter
                    //CR 18827
		    endIndex ++;
		    startIndex = endIndex;
                    //end CR 18827
		}
	    }
	    // check if this character is an escape character and the previous
	    // character is an escape character
	    else if ( mBOCharBuffer[endIndex] == mCharEscape &&
		mBOCharBuffer[endIndex - 1] == mCharEscape ) {

		// we found an escaped escape character, so save the portion
		// up to the first escape character in the buffer
		buffer.append(mBOCharBuffer, startIndex, endIndex -
		    mCurrentIndex);

		// move the startIndex to the character past the second
		// escape character
                //CR 15458
                endIndex ++;
		startIndex = endIndex;
	    }
	    else
		// this character is not a delimiter, and not an escaped escape
		// character so move to the next character
		endIndex++;
	}

	mCurrentIndex = endIndex + 1;
	return buffer.append(mBOCharBuffer, startIndex, endIndex - startIndex )
	    .toString();

    } // end charDelimiterRead

    /**
     *  Performs search for string delimiter and
     *  returns string before the delimter. Uses
     *  String.indexOf for searching.
     *
     *  @return         String value that is the
     *                  next token that appers in
     *                  the mBOCharBuffer.
     *
     *  Created for CR: CR12981
     */
    private String stringDelimiterRead()
    {
      String lvResult = null;
      StringBuffer lvResultBuffer = new StringBuffer(mAttrValueBufferSize);

      int lvDelimiterIndex = -1;
      int lvEscapIndex = -1;

      if (!charSubstring(mBOCharBuffer, mStartIndex, mStartIndex+mDelimiter.length()).equalsIgnoreCase(mDelimiter)
       && mStartIndex < mBOCharBuffer.length)
      {
	// This condition ensures non-empty token
	do
	{
	  // Check the data for the escape and delimiter characters

	  lvEscapIndex = charIndexOf(mBOCharBuffer, mEscape, mStartIndex);
	  lvDelimiterIndex = charIndexOf(mBOCharBuffer, mDelimiter, mStartIndex);

	  if (lvEscapIndex > -1 && lvEscapIndex < lvDelimiterIndex )
	  {
	    // Checks the character that appears in the string just after
	    //  the escape character.
	    String lvCharBehindEscape = charSubstring(mBOCharBuffer, lvEscapIndex + 1, lvEscapIndex + 1 + mDelimiter.length());

	    if ( lvCharBehindEscape.equalsIgnoreCase(mEscape) )
	    {
	      // Append the data up until the escape character appears, but
	      //  skip the next character.
	      lvResultBuffer.append(charSubstring(mBOCharBuffer, mStartIndex, lvEscapIndex));
	      lvResultBuffer.append(charSubstring(mBOCharBuffer, lvEscapIndex + 1, lvEscapIndex + 1 + mEscape.length()) );
	    }
	    else if ( lvCharBehindEscape.equalsIgnoreCase(mDelimiter) )
	    {
	      lvResultBuffer.append(charSubstring(mBOCharBuffer, mStartIndex, lvEscapIndex));
	      lvResultBuffer.append(charSubstring(mBOCharBuffer, lvEscapIndex + 1, lvEscapIndex + 1 + mDelimiter.length()) );
	    }
	    else
	      // This is the case that the escape character is part of the
	      //  actual data and is not being used as an escape character.
	      lvResultBuffer.append(charSubstring(mBOCharBuffer, mStartIndex, lvEscapIndex + mEscape.length() + 1));

	      mStartIndex = lvEscapIndex + mEscape.length() + 1;//ensured by the condition; it is OK because the end index is NOT inclusive

	  }
	  else
	  {
	    if (lvDelimiterIndex > -1)
	    {
	      lvResultBuffer.append(charSubstring(mBOCharBuffer, mStartIndex, lvDelimiterIndex));
	    }
	    else
	    {
	      lvResultBuffer.append(charSubstring(mBOCharBuffer, mStartIndex, mBOCharBuffer.length));
	    }

		      // Increment the start index to skip over the delimiter
	    mStartIndex = lvDelimiterIndex + mDelimiter.length();
	  }

	}

	while (lvEscapIndex > -1 && lvEscapIndex < lvDelimiterIndex && mStartIndex < mBOCharBuffer.length);
	lvResult = lvResultBuffer.toString();
	if ( mStartIndex >= mBOCharBuffer.length)
	mEndOfBOString = true;
      }

      else if (charSubstring(mBOCharBuffer, mStartIndex, mStartIndex + mDelimiter.length() ).equalsIgnoreCase(mDelimiter)  && mStartIndex < mBOCharBuffer.length)
      {
	// Define new starting search point for next token
	mStartIndex = mStartIndex + mDelimiter.length();
      }
      else
      {
	mEndOfBOString = true;
      }

      return lvResult;

    } // end stringDelimiterRead

    /**
     *  Returns the index within this charArray of the first occurrence of the
     *  specified substring, starting at the specified begin index.
     *
     *  There is no restriction on the value of <code>fromIndex</code>. If
     *  it is negative, it has the same effect as if it were zero: this entire
     *  string may be searched. If it is greater than the length of this
     *  string, it has the same effect as if it were equal to the length of
     *  this string: -1 is returned.
     *
     *  @param   charArray - the charArray to search in.
     *  @param   substring - the string to search for.
     *  @param   begin - the index to start the search from.
     *  @return  If the string argument occurs as a substring within this
     *          object at a starting index no smaller than
     *          <code>fromIndex</code>, then the index of the first character
     *          of the first such substring is returned. If it does not occur
     *          as a substring starting at <code>fromIndex</code> or beyond,
     *          -1 is returned.
     *  @exception java.lang.NullPointerException if mBOCharBuffer is
     *          null
     *  Created for CR: CR12981
     */
    private int charIndexOf(char[] charArray, String substring, int begin)
    {
	char v2[] = substring.toCharArray();
	int max = charArray.length - substring.length();

	if (begin >= charArray.length) {
	    if (charArray.length == 0 && begin == 0 && substring.length() == 0) {
		/* There is an empty string at index 0 in an empty string. */
		return 0;
	    }
	    /* Note: begin index might be near -1>>>1 */
	    return -1;
	}
	if (begin < 0) {
	    begin = 0;
	}
	if (substring.length() == 0) {
	    return begin;
	}

	char first  = v2[0];
	int i = begin;

    startSearchForFirstChar:
	while (true) {

	    /* Look for first character. */
	    while (i <= max && charArray[i] != first) {
		i++;
	    }
	    if (i > max) {
		return -1;
	    }

	    /* Found first character, now look at the rest of v2 */
	    int j = i + 1;
	    int end = j + substring.length() - 1;
	    int k = 1;
	    while (j < end) {
		if (charArray[j++] != v2[k++]) {
		    i++;
		    /* Look for str's first char again. */
		    continue startSearchForFirstChar;
		}
	    }
	    return i;	/* Found whole string. */
	}
    } // end charIndexOf

    /**
     *  Returns a new StringBuffer that is a substring of charArray. The
     *  substring begins with the character at the specified begin index and
     *  extends to the end index.
     *  Examples:
     *
     *  charArray = {'d', 'o', 'o', 'r'};
     *  charSubstring(charArray, 1, 2) returns "oo"
     *
     *  @param      charArray - the array to use for substring.
     *  @param      begin - the beginning index, inclusive.
     *  @param      end - the ending index, non-inclusive.
     *  @return     the specified substring.
     *  @exception  IndexOutOfBoundsException  if
     *              <code>beginIndex</code> is negative or larger than the
     *              length of this <code>String</code> object.
     *  Created for CR: CR12981
     */
    private String charSubstring(char[] charArray, int begin, int end)
    {
      StringBuffer returnString = new StringBuffer();

      for (int i = begin; i < end; i++)
      {
	returnString.append(charArray[i]);
      }

      return returnString.toString();
    } // end charSubstring


  /**
   * Parses the app-text of a business object
   * and generates a list of the attributes
   * reserved for child meta objects.
   *
   * @return list of meta object attribute names
   */
    public static List listMOAttrNames( BusinessObjectInterface busObj ) {
	List list = new LinkedList();
	StringTokenizer tok = new StringTokenizer( busObj.getAppText(), ";" );
	while( tok.hasMoreTokens() ) {
	    String nextToken = tok.nextToken();
	    try {
		int nameValueDelimPosition = nextToken.indexOf( '=' );
		String tag = nextToken.substring( 0, nameValueDelimPosition ).trim().toLowerCase();
		String attributeName = nextToken.substring( nameValueDelimPosition + 1 ).trim();
		if( tag.startsWith( "cw_mo_" ) ) {
		    list.add( attributeName );
		}
	    }
	    catch( Exception e ) {
	      //ignore non-name-value pair elements
	    }
	}
	return list;
    }//method listMOAttrNames
}
