/* IBM Confidential
 * OCO Source Materials
 * 5724-C11, 5724-D17,5724-E30
 * (C) Copyright IBM Corporation 1997,2002
 * The source code for this program is not published or otherwise
 * divested of its trade secrets, irrespective of what has been
 * deposited with the U.S. Copyright Office.
*/

package com.crossworlds.DataHandlers.text;

import java.util.List;
import java.util.LinkedList;
import java.util.StringTokenizer;

import com.crossworlds.DataHandlers.DataHandler;
import com.crossworlds.DataHandlers.Exceptions.CW_BOFormatException;
import com.crossworlds.DataHandlers.Exceptions.CW_EndOfBOException;
import com.crossworlds.DataHandlers.Exceptions.CW_SubscriptionException;

//CW classes
import CxCommon.Exceptions.CxObjectNoSuchAttributeException;
import CxCommon.Exceptions.CxObjectInvalidAttrException;
import CxCommon.Exceptions.BusObjSpecNameNotFoundException;
import CxCommon.BusinessObjectInterface;
import CxCommon.CxObjectContainerInterface;
import CxCommon.CxObjectAttr;
import AppSide_Connector.JavaConnectorUtil;

// java classes
import java.io.InputStream;
import java.io.ByteArrayInputStream;
import java.io.Reader;

/**
*   Implementation of DataHandler base class desinged to convert between a
*   serialized business object in fixed width format and a business object.
*   Assumes the business object name and verb will be included in the data for
*   each object as well as the count for child objects.
*/
public class fixedwidth extends DataHandler
{
    protected final static String AlignmentBoth = "BOTH";
    protected final static String AlignmentLeft = "LEFT";
    protected final static String AlignmentRight = "RIGHT";

    protected final int mBufferSize = 2048;  //2K Characters
    protected final int mAttrValueBufferSize = 256;

    // CR9251 - Removed all of the transient keywords
    protected String mCxIgnore;
    protected String mCxBlank;
    protected String mAlignment;
    protected boolean mTruncation;
    protected char mPadCharacter;
    protected int mBONameSize;
    protected int mBOVerbSize;
    protected int mBOCountSize;

    //for the BOString parsing
    protected int mStartIndex;  //of a line
    protected int mTokenNum;     //for error messaging
    protected int mBOStringLength;
    protected boolean mEndOfBOString;  //set in this.getNextLine()

    protected String mBOString;
    protected StringBuffer mBOStringBuffer;

    public fixedwidth()
    {

    }

    /**
    *   Set the configuration options based on the options set through the
    *   base class.
    */
    private void getConfigOptions()
    {
        traceWrite( "Entering getConfigOptions()", JavaConnectorUtil.LEVEL4);

        // Set the member variables according to the properties set
        //  based on the config object.
        mCxIgnore = getOption("CxIgnore");
        mCxBlank = getOption("CxBlank");
        mAlignment = getOption("Alignment");

        String lvTruncation = getOption("Truncation");
        if( lvTruncation.compareTo("true") == 0 )
            mTruncation = true;
        else if( lvTruncation.compareTo("false") == 0 )
            mTruncation = false;
        else
            mTruncation = true;

        String lvPadCharacter = getOption("PadCharacter");
        if( lvPadCharacter.length() > 0 )
            mPadCharacter = lvPadCharacter.charAt(0);
        else
            mPadCharacter = ' ';

        String lvConfigInt = null;
        Integer lvIntFromConfig = null;

        lvConfigInt = getOption("BONameSize");
        if( lvConfigInt.length() > 0 )
        {
            lvIntFromConfig = new Integer(lvConfigInt);
            mBONameSize = lvIntFromConfig.intValue();
        }
        else
            mBONameSize = -1;

        lvConfigInt = null;
        lvConfigInt = getOption("BOCountSize");
        if( lvConfigInt.length() > 0 )
        {
            lvIntFromConfig = new Integer(lvConfigInt);
            mBOCountSize = lvIntFromConfig.intValue();
        }
        else
            mBOCountSize = -1;

        lvConfigInt = null;
        lvConfigInt = getOption("BOVerbSize");
        if( lvConfigInt.length() > 0 )
        {
            lvIntFromConfig = new Integer(lvConfigInt);
            mBOVerbSize = lvIntFromConfig.intValue();
        }
        else
            mBOVerbSize = -1;

        traceWrite( "Exiting getConfigOptions", JavaConnectorUtil.LEVEL4);
    }

    /**
    *   Resets the value of all memeber variables that should only have the
    *   lifespan of one conversion between serialized data and a business
    *   object.
    */
    private void clear(Object config) throws Exception
    {
        traceWrite( "Entering clear()", JavaConnectorUtil.LEVEL4);

        if ( config instanceof BusinessObjectInterface )
        {
            traceWrite( "Setting options based on config object attributes",
                        JavaConnectorUtil.LEVEL4);
            setupOptions( (BusinessObjectInterface)config);
        }

        getConfigOptions();

        mBOString = null;
        BusinessObjectInterface mBO = null;
        mBOStringBuffer = new StringBuffer(mBufferSize);

//        if( mCxIgnore.length() == 0)
//            mCxIgnore = "CxIgnore";
//        if( mCxBlank.length() == 0)
//            mCxBlank = "CxBlank";
        if( mAlignment.length() == 0)
            mAlignment = "both";

        if( mBONameSize < 0 )
            mBONameSize = 30;
        if( mBOVerbSize < 0 )
            mBOVerbSize = 20;
        if( mBOCountSize < 0 )
            mBOCountSize = 10;

        mStartIndex = 0;
        mTokenNum = 0;
        mBOStringLength = 0;
        mEndOfBOString = false;

        traceWrite( "Exiting clear()", JavaConnectorUtil.LEVEL4);
    }

    /**
    *   Converts the data contained in a Reader object to a CrossWorlds
    *   business object.  Allows configuration options to be set by passing
    *   in a meta object which contains the config information.
    *   @param  serializedData  Reader object which contains the data to be
    *                           converted to a business object.
    *   @param  config          If this parameter is an instance of
    *                           BusinessObjectInterface config options will
    *                           be set based on the default values of the
    *                           attributes.
    */
    public BusinessObjectInterface  getBO(Reader serializedData, Object config)
        throws Exception
    {
        traceWrite( "Entering getBO(Reader, Object)", JavaConnectorUtil.LEVEL4);

        clear(config);

        BusinessObjectInterface resultObj = null;

        // Create a String object from the Reader, then use the stirng based
        //  method
        int conversionCheck;
        char[] temp = new char[2000];
        StringBuffer tempStringBuffer = new StringBuffer(1000);

        while ( (conversionCheck = serializedData.read (temp)) != -1)
            tempStringBuffer.append(new String (temp, 0, conversionCheck));

        mBOString = new String(tempStringBuffer);
        traceWrite( "getBO(Reader, Object) set mBOString to the following: \n"
                    + mBOString, JavaConnectorUtil.LEVEL5);
        mBOStringLength = mBOString.length();


        traceWrite("getBO(Reader, Object) calling getBOFromString",
                    JavaConnectorUtil.LEVEL5);
        resultObj = getBOFromString(null);

		// <I18N> by Chen Gang (07/22/2002)
		if(resultObj != null && getLocale() != null)
			resultObj.setLocale(getLocale());
		// </I18N>
        traceWrite( "Exiting getBO(Reader, Object)", JavaConnectorUtil.LEVEL4);

        return resultObj;
    }

    /**
    *   Populates a given BusinessObjectInterface with data extracted
    *   from Reader object. In this case it is up to the caller to create
    *   the BO.  Allows configuration options to be set by passing in a
    *   meta object which contains the config information.
    *   @param  serializedData  Reader object which contains the data to
    *                           be converted to a business object.
    *   @param  theObj          BusinessObjectInterface to be populated
    *                           with data from Reader object
    *                           'serializedData'.
    *   @param  config          If this parameter is an instance of
    *                           BusinessObjectInterface config options will
    *                           be set based on the default values of the
    *                           attributes.
    */
    public void getBO(Reader serializedData, BusinessObjectInterface theObj,Object config)
    throws Exception
    {
        traceWrite( "Entering getBO(Reader, BusinessObjectInterface, Object) for BO type "
                    + theObj.getName(), JavaConnectorUtil.LEVEL4);

        clear(config);

        // Create a String object from the Reader, then use the strinng based
        //  method
        int conversionCheck;
        char[] temp = new char[2000];
        StringBuffer tempStringBuffer = new StringBuffer(1000);

        while ( (conversionCheck = serializedData.read (temp)) != -1)
            tempStringBuffer.append(new String (temp, 0, conversionCheck));

        mBOString = new String(tempStringBuffer);
        mBOStringLength = mBOString.length();
        traceWrite( "getBO(Reader, BusinessObjectInterface, Object) set mBOStringLength to the following: \n"
                    + mBOStringLength, JavaConnectorUtil.LEVEL5);

        getBOFromString( theObj );

		// <I18N> by Chen Gang (07/22/2002)
		if(theObj != null && getLocale() != null)
			theObj.setLocale(getLocale());
		// </I18N>

        traceWrite( "Exiting getBO(Reader, BusinessObjectInterface, Object) for BO type "
                    + theObj.getName(), JavaConnectorUtil.LEVEL4);

    }

    /**
    *   Populates a given BusinessObjectInterface with data extracted
    *   from String object. In this case it is up to the caller to create
    *   the BO.  Allows configuration options to be set by passing in a meta
    *   object which contains the config information.
    *   @param  serializedData  String object which contains the data to be
    *                           converted to a business object.
    *   @param  theObj          BusinessObjectInterface to be populated with
    *                           data from Reader object 'serializedData'.
    *   @param  config          If this parameter is an instance of
    *                           BusinessObjectInterface config options will be
    *                           set based on the default values of the
    *                           attributes.
    */
    public void getBO(String serializedData, BusinessObjectInterface theObj,
                    Object config )
        throws Exception
    {
        traceWrite( "Entering getBO(String, BusinessObjectInterface, Object)",
                    JavaConnectorUtil.LEVEL4);

        clear(config);

        mBOString = new String(serializedData);
        mBOStringLength = serializedData.length();

        traceWrite("getBO(String, BusinessObjectInterface, Object) calling getBOFromString",
                    JavaConnectorUtil.LEVEL5);
        getBOFromString( theObj );

		// <I18N> by Chen Gang (07/22/2002)
		if(theObj != null && getLocale() != null)
			theObj.setLocale(getLocale());
		// </I18N>

        traceWrite( "Exiting getBO(String, BusinessObjectInterface, Object)",
                    JavaConnectorUtil.LEVEL4);
    }

    /**
    *   Creates an instance of BusinessObjectInterface and populates the verb.
    *   Calls parseAttributeList to do the parsing of the String and populate
    *   the attribute values.  Determines what type of business object to
    *   create either based on the paramter value or on the first token in the
    *   data.  The paramter value must match the first token in the data,
    *   otherwise an error is generated.
    *   @param  pvBOType    This is the type of Business Object that should be
    *                       created.  If this is not null it is checked against
    *                       the first token in the data, and an error is
    *                       generated if there is a mis-match.
    */
    private BusinessObjectInterface getBOFromString(BusinessObjectInterface pvBO)
        throws Exception
    {
        traceWrite("Entering getBOFromString(String)",
                    JavaConnectorUtil.LEVEL4);

        String pvBOType = null;

        if( pvBO != null )
        {
            pvBOType = pvBO.getName();
        }

        String lvBOName = null;
        String lvVerb = null;

        lvBOName = this.getNextToken(mBONameSize, true);
        lvVerb = this.getNextToken(mBOVerbSize, true);

        if( (pvBOType != null) && (lvBOName.compareTo(pvBOType) != 0) )
        {
            throw new Exception("Name from input source (" + lvBOName
                                + ") doesn't match type " + pvBOType
                                + " passed to getBOFromString");
        }
        else if(pvBO == null)
        {
            pvBO = JavaConnectorUtil.createBusinessObject(lvBOName);
        }

        if (! mCxIgnore.equals( lvVerb) && ! mCxBlank.equals(lvVerb) )
        {
            pvBO.setVerb(lvVerb);
        }


        parseAttributeList(pvBO);

        traceWrite( "Exiting getBOFromString(String) with BO type " + pvBOType,
                    JavaConnectorUtil.LEVEL4);

        return pvBO;
    }

    /**
    *   Converts the data in a business object to a fixed length field
    *   String representation.
    *   @param  theObj      BusinessObjectInterface which contains the data
    *                       with which to build the string.  The spec
    *                       associated with this business object should have
    *                       the MaxLength property accurately set for all the
    *                       attributes, as this is what the format of the
    *                       String is based on.
    *   @param  config      If this parameter is an instance of
    *                       BusinessObjectInterface config options will be
    *                       set based on the default values of the
    *   @return     String representation of the data in the business object.
    *               Format is fixed width fields.
    */
    public String getStringFromBO(BusinessObjectInterface theObj, Object config)
        throws Exception
    {
        traceWrite( "Entering getStringFromBO(BusinessObjectInterface, Object) "
                    + " for object type " + theObj.getName(),
                    JavaConnectorUtil.LEVEL4);

        clear(config);

        String lvBOString = null;

        setAttrList(theObj);

        lvBOString = mBOStringBuffer.toString();

        traceWrite( "Exiting getStringFromBO(BusinessObjectInterface, Object) "
                    + " for object type " + theObj.getName(),
                    JavaConnectorUtil.LEVEL4);

        return lvBOString;
    }

    /**
    *   Converts the data in a business object to a fixed length field
    *   InputStream representation.  This is done by first building a String
    *   based on the business object data, then converting that String to an
    *   InputStream.
    *   @param  theObj      BusinessObjectInterface which contains the data
    *                       with which to build the stream.  The spec
    *                       associated with this business object should have
    *                       the MaxLength property accurately set for all the
    *                       attributes, as this is what the format of the
    *                       String is based on.
    *   @param  config      If this parameter is an instance of
    *                       BusinessObjectInterface config options will be
    *                       set based on the default values of the
    *   @return     InputStream representation of the data in the business object.
    *               Format is fixed width fields.
    */
    public InputStream getStreamFromBO(BusinessObjectInterface theObj, Object config)
        throws Exception
    {
        traceWrite( "Entering getStreamFromBO(BusinessObjectInterface, Object) "
                    + " for object type " + theObj.getName(),
                    JavaConnectorUtil.LEVEL4);

        clear(config);

        String BOstring;

        BOstring = getStringFromBO(theObj, config);

        traceWrite( "Exiting getStreamFromBO(BusinessObjectInterface, Object) "
                    + " for object type " + theObj.getName(),
                    JavaConnectorUtil.LEVEL4);

		// <I18N> by Chen Gang (07/22/2002)
		// return new ByteArrayInputStream( BOstring.getBytes() );
		if(getEncoding() != null) {
			return new ByteArrayInputStream( BOstring.getBytes(getEncoding()));
		} else {
			return new ByteArrayInputStream( BOstring.getBytes());
		}
		// </I18N>
    }

    /**
    *   Parses the next token from mBOString based on the length of
    *   the token.  Also trims the field of whitespace.
    *   @param  pvTokeLength    The length (number of characters) of the
    *                           next token.
    *   @param  pvTrimBoth      If 'true' passed in, then the token is
    *                           trimmed on both sides.  If 'false' is passed,
    *                           it is only trimmed according to the option
    *                           set.
    *   @return         String value that is the next token that appers in the
    *                   mBOString.
    */
    protected String getNextToken(int pvTokeLength, boolean pvTrimBoth)
        throws Exception
    {
        traceWrite( "Entering getNextToken(int, boolean) with values " + pvTokeLength
                    + " and " + pvTrimBoth + ".  Start index for mBOString is "
                    + mStartIndex, JavaConnectorUtil.LEVEL4);

        if (mStartIndex >= mBOStringLength || mEndOfBOString)
            throw new Exception(" Already reached end of BO while expecting more token ");

        String lvResultString = null;
        if (pvTokeLength > 0)  //else no sense to do anything
        {
            int lvLeftIndex = mStartIndex;
            int lvRightIndex = mStartIndex + pvTokeLength ;

            if (lvRightIndex > mBOStringLength)
                lvRightIndex = mBOStringLength;  // take care of the last token

            //trim according to configuration
            if (pvTrimBoth)
            {
                lvLeftIndex = this.trimLeft(lvLeftIndex, lvRightIndex);
                lvRightIndex = this.trimRight(lvRightIndex, lvLeftIndex);
            }
            else
            {
                if (mAlignment.equals(fixedwidth.AlignmentBoth))
                {
                    lvLeftIndex = this.trimLeft(lvLeftIndex, lvRightIndex);
                    lvRightIndex = this.trimRight(lvRightIndex, lvLeftIndex);
                }
                else if (mAlignment.equals(fixedwidth.AlignmentLeft))
                    lvRightIndex = this.trimRight(lvRightIndex, lvLeftIndex);
                else if (mAlignment.equals(fixedwidth.AlignmentRight))
                    lvLeftIndex = this.trimLeft(lvLeftIndex, lvRightIndex);
            }

            lvResultString = mBOString.substring(lvLeftIndex, lvRightIndex);
            mStartIndex += pvTokeLength;
            if (mStartIndex > mBOStringLength)
            mEndOfBOString = true;
        }

        // This may constitute a performance problem if the result strings are
        //  very large, however, in most cases a single token should be relatively
        //  small.
        traceWrite( "Exiting getNextToken(int, boolean) with values "
                    + lvResultString, JavaConnectorUtil.LEVEL4);

        return lvResultString;
    }

    /**
    *   Parses mBOString to populate the attributes for the
    *   BusinessObjectInterface.
    *   @param  pvBO        BusinessObjectInterface to populate with the
    *                       data parsed from the string representation.
    */
    protected void  parseAttributeList(BusinessObjectInterface pvBO)
        throws Exception
    {
        traceWrite( "Entering parseAttributeList(BusinessObjectInterface) for bo type "
                    + pvBO.getName(), JavaConnectorUtil.LEVEL4);

        if ( mEndOfBOString )
            throw new Exception(" Col " + mStartIndex + ":  immature ending of BO.  Please check your data format; ");
        else if( pvBO == null )
            throw new Exception(" parseAtributeList received null BusinessObjectInterface");

        int lvAttrNum = 0;
        if (getOption("OmitObjectEventId").equalsIgnoreCase("true"))
        {
          lvAttrNum = pvBO.getAttrCount()-1; //skip objectEventId
        }
        else
        {
          lvAttrNum = pvBO.getAttrCount();
        }

        String lvAttrName = null;
        String lvAttrValue = null;
        int lvAttrMaxLength = 0;

        try
        {
            List moAttrNames = listMOAttrNames( pvBO );

            ATTRIBUTE_WALK: for (int lvAttrIndex = 0; lvAttrIndex < lvAttrNum; lvAttrIndex++)
            {
                CxObjectAttr lvAttrSpec = pvBO.getAttrDesc(lvAttrIndex);
                lvAttrName = lvAttrSpec.getName();

                //skip child objects designated as meta objects
                if( moAttrNames.contains( lvAttrName ) ) {
                    continue ATTRIBUTE_WALK;
                }

                // Check to see if the attribute is a simple attribute,
                //  or is a business object container.
                if (lvAttrSpec.isObjectType())
                {
                    // Get the next token based on the BOCountSize
                    lvAttrMaxLength = mBOCountSize;
                    lvAttrValue =  this.getNextToken(mBOCountSize, true);
                    String lvBOType = lvAttrSpec.getTypeName();
                    Object lvAttrBOValue = null;

                    if (lvAttrSpec.isMultipleCard())
                    {
                        traceWrite( "parseAttributeList found multiple cardinality container for "
                                    + lvBOType,
                                    JavaConnectorUtil.LEVEL5);
                        this.getMultipleCard(pvBO, lvAttrIndex, lvBOType, lvAttrValue);
                    }
                    else
                    {
                        traceWrite( "parseAttributeList found single cardinality container for "
                                    + lvBOType,
                                    JavaConnectorUtil.LEVEL5);
                        this.getSingleCard(pvBO, lvAttrIndex, lvBOType, lvAttrValue);
                    }

                }
                else
                {
                    // Get the next token based on the MaxLength of the attribute
                    lvAttrMaxLength = lvAttrSpec.getMaxLength();
                    if (lvAttrMaxLength > 0)
                        lvAttrValue =  this.getNextToken(lvAttrMaxLength, false);
                    else
                        lvAttrValue = null;

                    // For simple (String) attribute, we check to see if the token is either
                    //  null, or matches the configured CxIgnore or CxBlank values.
                    //  Otherwise, set the attribute value with the token value
                    if (lvAttrValue == null )
                        pvBO.setAttrValue(lvAttrIndex, null);   //this includes ObjectEventIdAttr
                    else if (lvAttrValue.equals(mCxIgnore) || lvAttrValue.equals("") )
                        pvBO.setAttrValue(lvAttrIndex, null);   //this includes ObjectEventIdAttr
                    else if (lvAttrValue.equals(mCxBlank) || lvAttrValue.equals(" "))
                        pvBO.setAttrValue(lvAttrIndex, "");   //this includes ObjectEventIdAttr
                    else
                        pvBO.setAttrValue(lvAttrIndex, lvAttrValue);   //this includes ObjectEventIdAttr
                }
            }
        }
        catch (CxObjectNoSuchAttributeException e)
        {
            //found format error of this BO
            throw new Exception(e.toString() + ": data format error; Attribute miss alignment at: attribute name = " +
            lvAttrName + "; column = " + getStartIndex(lvAttrMaxLength) );
        }

        traceWrite( "Exiting parseAttributeList(BusinessObjectInterface) for bo type "
                    + pvBO.getName(), JavaConnectorUtil.LEVEL4);
    }

    /**
    *   Populates a child container based on token values in the fixedwidth
    *   string.
    *   Expects all objects in the child container to be of the same type.
    *   @param  pvParentBO              Parent business object.
    *   @param  pvParentAttrIndex       Attribute index of the child to be
    *                                   created.
    *   @param  pvBOType                Type of the child business object.
    *   @param  pvObjectCountString     Number of instances of the child
    *                                   object as parsed from the delimited
    *                                   string.
    */
    protected void getMultipleCard(BusinessObjectInterface pvParentBO,
                                   int pvParentAttrIndex, String pvBOType,
                                   String pvObjectCountString)
        throws CW_BOFormatException, Exception
    {
        traceWrite( "Entering getMultipleCard for parentBO "
                    + pvParentBO.getName() + " index " + pvParentAttrIndex
                    + " with count " + pvObjectCountString,
                    JavaConnectorUtil.LEVEL4);

        try
        {
            if ( pvObjectCountString.equals(mCxIgnore) )
            {
                traceWrite( "Count string for container " + pvBOType
                            + " in parent " + pvParentBO.getName()
                            + " is CxIgnore value",
                            JavaConnectorUtil.LEVEL5);
            }
            else
            {
                int lvObjectCount = Integer.parseInt(pvObjectCountString);

                if ( lvObjectCount == 0)
                {
                    traceWrite( "Count string for container " + pvBOType
                                + " in parent " + pvParentBO.getName()
                                + " is 0",
                                JavaConnectorUtil.LEVEL5);
                }
                else if (lvObjectCount > 0)
                {
                    // There is at least one instance of the object in the
                    //  delimited string

                    traceWrite( "Found " + lvObjectCount
                                + " instances of child object pvBOType "
                                + " in parent object " + pvParentBO.getName(),
                                JavaConnectorUtil.LEVEL5);

                    BusinessObjectInterface lvChildBO = JavaConnectorUtil.createBusinessObject(pvBOType);
                    // For each instance of the child object, parse the
                    //  attribute list, and then add the object container
                    //  to the parent.
                    for (int lvObjectIndex = 0; lvObjectIndex < lvObjectCount; lvObjectIndex++)
                    {
                        lvChildBO = getBOFromString(null);
                        pvParentBO.setAttrValue(pvParentAttrIndex, lvChildBO);
                    }
                }
            }
        }
        catch (CxObjectNoSuchAttributeException e)
        {
            throw new CW_BOFormatException(" Column: " + getStartIndex(mBOCountSize) +
                                        ": formatting error; " + e.getMessage() + ". ");
        }
        catch (CxObjectInvalidAttrException e)
        {
            throw new CW_BOFormatException(" Column: " + getStartIndex(mBOCountSize) +
                ": formatting error: Attribute miss alignment; " + e.getMessage() + ". ");
        }
        catch (NumberFormatException e)
        {
            throw new CW_BOFormatException(" Column: " + getStartIndex(mBOCountSize) +
                ": formatting error:  cannot recognize number format; " + e.getMessage() + ". ");
        }
        /*
        catch (BusObjSpecNameNotFoundException e)
        {
            throw new CW_BOFormatException("Invalid BusObjSpec name found: " + e.getMessage());
        }
        */

        traceWrite( "Exiting getMultipleCard for parentBO " + pvParentBO.getName() + " index " + pvParentAttrIndex
                    + " with count " + pvObjectCountString,
                    JavaConnectorUtil.LEVEL4);

    }

    /**
    *   Populates a single cardinatlity child business object with the token
    *   values found in the fixed width string.
    *   @param  pvParentBO              Parent business object.
    *   @param  pvParentAttrIndex       Attribute index of the child to be
    *                                   created.
    *   @param  pvBOType                Type of the child business object.
    *   @param  pvObjectCountString     Number of instances of the child
    *                                   object as parsed from the delimited
    *                                   string.
    */
    protected void  getSingleCard(BusinessObjectInterface pvParentBO,
                                  int pvParentAttrIndex, String pvBOType,
                                  String pvObjectCountString)
        throws CW_BOFormatException, Exception
    {
        traceWrite( "Entering getSingleCard for parentBO "
                    + pvParentBO.getName() + " index " + pvParentAttrIndex
                    + " with count " + pvObjectCountString,
                    JavaConnectorUtil.LEVEL4);

        try
        {
            BusinessObjectInterface lvChildBO = null;

            // Check the object count token
            // If it does not match "1", assume that the child object should
            //  be null
            if (pvObjectCountString.equals("1"))
            {
                // The delimited string contains a single instance of the
                //  child object
                traceWrite( "Found single instance of child object "
                            + pvBOType + " in delimited string",
                            JavaConnectorUtil.LEVEL5);

                lvChildBO = JavaConnectorUtil.createBusinessObject(pvBOType);
                getBOFromString(lvChildBO);
                pvParentBO.setAttrValue(pvParentAttrIndex, lvChildBO);
            }
            else if ( pvObjectCountString.equals(mCxIgnore) ||
                        pvObjectCountString.equals("0"))
            {
                // This case is included to validate the object count token is valid
                traceWrite( "Encountered null child object " + pvBOType
                            + " for parent " + pvParentBO.getName(),
                            JavaConnectorUtil.LEVEL5);
            }
            else
                throw new CW_BOFormatException(" Column: "
                        + getStartIndex(mBOCountSize) +
                        " BO object count error.  This count should be 1, 0, or CxIgnore. " );

        }
        catch (CxObjectInvalidAttrException e)
        {
            throw new CW_BOFormatException(" Column: " +
                                            getStartIndex(mBOCountSize) + ": try to access to the wrong attribute. " +  e.getMessage() + ". ");
        }
        catch (CxObjectNoSuchAttributeException e)
        {
            throw new CW_BOFormatException(" Column: " + getStartIndex(mBOCountSize) +
                                            ": try to access to the wrong attribute. " +  e.getMessage() + ". ");
        }
        /*
        catch (BusObjSpecNameNotFoundException e)
        {
            throw new CW_BOFormatException("Invalid BusObjSpec name found: "
                                + e.getMessage());
        }
        */

        traceWrite( "Exiting getSingleCard for parentBO " + pvParentBO.getName() + " index " + pvParentAttrIndex
                    + " with count " + pvObjectCountString,
                    JavaConnectorUtil.LEVEL4);
    }

    /**
    *   Loops through the attributes in a BusinssObjectInterface and add the
    *   attribute values to the mBOStringBuffer.
    *   @param  pvBO    BusinessObjectInterface which contains the data to
    *                   be added to the fixed width String.
    */
    protected void setAttrList(BusinessObjectInterface pvBO) throws Exception
    {
        traceWrite( "Entering setAttrList(BusinessObjectInterface) for object "
                   + pvBO.getName(), JavaConnectorUtil.LEVEL4);

        String lvAttrName = null;
        String lvAttrValue = null;
        int lvAttrMaxLength = 0;

        // Add the business object name and verb to the fixed width format
        //  String
        this.setSimpleToken( mBONameSize, pvBO.getName());
        this.setSimpleToken( mBOVerbSize, pvBO.getVerb());

        try
        {
            List moAttrNames = listMOAttrNames( pvBO );

            int lvAttrCount = 0;
            if (getOption("OmitObjectEventId").equalsIgnoreCase("true"))
            {
              lvAttrCount = pvBO.getAttrCount()-1; //skip objectEventId
            }
            else
            {
              lvAttrCount = pvBO.getAttrCount();
            }

            ATTRIBUTE_WALK: for (int lvAttrIndex = 0; lvAttrIndex < lvAttrCount; ++lvAttrIndex)
            {
                CxObjectAttr lvAttrSpec = pvBO.getAttrDesc(lvAttrIndex);
                lvAttrName = lvAttrSpec.getName();

                // Check if the current attribute is a simple (String) type
                //  or a contained object.
                if (lvAttrSpec.isObjectType())
                {
                    //skip child objects designated as meta objects
                    if( moAttrNames.contains( lvAttrName ) ) {
                      continue ATTRIBUTE_WALK;
                    }

                    if (lvAttrSpec.isMultipleCard())
                    {
                        CxObjectContainerInterface lvAttrMultiCardBOValue = (CxObjectContainerInterface) pvBO.getAttrValue(lvAttrIndex);

                        if (lvAttrMultiCardBOValue == null)
                        {
                            traceWrite( "setAttrList found empty multiple cardinality container " + lvAttrSpec.getTypeName(),
                                        JavaConnectorUtil.LEVEL5);
                            // Add the count to the fixed width String
                            this.setSimpleToken( mBOCountSize, "0");
                        }
                        else
                        {
                            int lvObjectCount = lvAttrMultiCardBOValue.getObjectCount();
                            traceWrite( "setAttrList found multiple cardinality container " + lvAttrSpec.getTypeName()
                                        + " with " + lvObjectCount + " instances",
                                        JavaConnectorUtil.LEVEL5);
                            // Add the count to the fixed width String
                            this.setSimpleToken( mBOCountSize, Integer.toString(lvObjectCount));

                            // Add each object in the container to the fixed
                            //  width String.
                            for (int lvContObjectIndex = 0; lvContObjectIndex < lvObjectCount; ++lvContObjectIndex)
                                setAttrList(lvAttrMultiCardBOValue.getBusinessObject(lvContObjectIndex));
                        }
                    }
                    else
                    {

                        BusinessObjectInterface lvAttrSingleCardBOValue = (BusinessObjectInterface) pvBO.getAttrValue(lvAttrIndex);

                        if (lvAttrSingleCardBOValue == null)
                        {
                            traceWrite( "setAttrList found empty single cardinality container " + lvAttrSpec.getTypeName(),
                                        JavaConnectorUtil.LEVEL5);
                            // Add the count to the fixed width String
                            this.setSimpleToken( mBOCountSize, "0");
                        }
                        else
                        {
                            traceWrite( "setAttrList found single cardinality container " + lvAttrSpec.getTypeName(),
                                        JavaConnectorUtil.LEVEL5);
                            // Add the count to the fixed width String
                            this.setSimpleToken( mBOCountSize, "1");
                            setAttrList(lvAttrSingleCardBOValue);
                        }
                    }
                }
                else
                {
                    lvAttrValue = (String) pvBO.getAttrValue(lvAttrIndex);
                    lvAttrMaxLength = lvAttrSpec.getMaxLength();

                    if (lvAttrMaxLength > 0)
                        this.setSimpleToken(lvAttrMaxLength, lvAttrValue);
                }
            }
        }
        catch (CxObjectNoSuchAttributeException e)
        {
            throw new Exception(e.getMessage());
        }

        traceWrite( "Exiting setAttrList(BusinessObjectInterface) for object "
                    + pvBO.getName(), JavaConnectorUtil.LEVEL4);

    }


    /**
    *   Adds a token value with correct pad char if needed to the data String
    *   being created (mBOStringBuffer).  Also contains the logic to check
    *   for the escape and delimiter character in the token value passed in
    *   if needed.
    *   @param  pvCellSize      Number of characters to add to this String for
    *                           this token.
    *   @param  pvTokenValue    Characters to be added to the String for this
    *                           token.
    */
    protected void setSimpleToken( int pvCellSize, String pvTokenValue)
        throws Exception
    {
        traceWrite( "Entering setSimpleToken(int, String)",
                    JavaConnectorUtil.LEVEL4);

        StringBuffer lvNewBuffer = new StringBuffer(pvCellSize);
        int lvTokenLength = 0;
        int lvCxIgnoreLength = mCxIgnore.length();
        int lvCxBlankLength = mCxBlank.length();

        int lvPadNumber = 0;

        // Check the token value to see if it is null
        if (pvTokenValue == null)
        {
            // For this case, we add the configured CxIgnore value to the
            //  fixed width String if it fits in the cell.
            if (!mTruncation && lvCxIgnoreLength > pvCellSize)
                throw new Exception(" Null attribute value encountered where cell size is "
                                + pvCellSize + ", size of CxIgnore value is "
                                + lvCxIgnoreLength + " and trucation is not allowed.  Please check your MO format configuration. ");
            else
            {
                lvPadNumber = pvCellSize - lvCxIgnoreLength;
                lvNewBuffer.append(mCxIgnore);
            }
        }
        else if (pvTokenValue.equals(""))
        {
            // For this case, the configured CxBlank value is added to the
            //  fixed width String if it fits in the cell.
            if (! mTruncation && lvCxBlankLength > pvCellSize)
                throw new Exception(" Blank attribute value encountered where cell size is "
                                + pvCellSize + ", size of CxBlank value is "
                                + lvCxBlankLength + " and trucation is not allowed.  Please check your MO format configuration. ");
            else
            {
                lvPadNumber = pvCellSize - lvCxBlankLength;
                lvNewBuffer.append(mCxBlank);
            }
        }
        else
        {
            // For this case, actually add the token value to the fixed width
            //  String, unless the data is too long for the cell.
            lvTokenLength = pvTokenValue.length();
            if (!mTruncation && lvTokenLength > pvCellSize )
                throw new Exception(" Attribute value encountered where cell size is "
                                + pvCellSize + ", size of token value is "
                                + lvTokenLength + " and trucation is not allowed.  Please check your MO format configuration. ");
            else
            {
                lvNewBuffer.append(pvTokenValue);
                lvPadNumber = pvCellSize - lvTokenLength;
            }
        }


        if (lvPadNumber <= 0 && mTruncation)
            // Token is longer than the cell and truncation is allowed,
            //  so the characters up to pvCellSize are added
            lvNewBuffer.setLength(pvCellSize);
        else if (lvPadNumber > 0)
        {
            // Pad the cell based on the configuration option chosen
            if ( mAlignment.equals(fixedwidth.AlignmentLeft) ||
                 mAlignment.equals(fixedwidth.AlignmentBoth))
                this.padRight(lvNewBuffer, lvPadNumber);
            else if (mAlignment.equals(fixedwidth.AlignmentRight))
                this.padLeft(lvNewBuffer, lvPadNumber);
        }

        String lvNewBuffString =  lvNewBuffer.toString();

        // Note that this may cause a performance issue when the tracing
        //  level is low, but in most cases it should not as any one token
        //  is *usually* not very long.
        traceWrite( "Adding the following token to the fixed width String: "
                    + lvNewBuffString, JavaConnectorUtil.LEVEL5);

        // After the cell has been fully formatted, append to fixed width
        //  String being built
        mBOStringBuffer.append(lvNewBuffString);

        traceWrite( "Exiting setSimpleToken(int, String)",
                    JavaConnectorUtil.LEVEL4);
    }


    /**
    *   Computes start index of current token
    *   @param  pvTokenSize     Number of characters in current token
    *   @return     Starting index of the current token calculated by
    *               subtracting the current cell size from the start index
    *               of the next cell.
    */
    protected int getStartIndex(int pvTokenSize)
    {
        return (mStartIndex - pvTokenSize);
    }

    /**
    *   Compute start of token data in mBOString by checking for
    *   pad characters starting on the left of the cell and moving
    *   right.
    *   @param  pvStartIndex    Start index of the cell of interest
    *                           in mBOString
    *   @param  pvAttrEndIndex  End index of the cell of interest in
    *                           mBOString
    *   @return     Starting index of data in cell of interest.
    */
    protected int trimLeft(int pvStartIndex, int pvAttrEndIndex)
    {
        int lvLeftIndex = pvStartIndex;
        while (mPadCharacter == mBOString.charAt(lvLeftIndex) &&
                lvLeftIndex < mBOStringLength &&
                lvLeftIndex < pvAttrEndIndex)
            lvLeftIndex++;
        return lvLeftIndex;
    }

    /**
    *   Compute start of token data mBOString by checking for pad
    *   characters starting on the right of the cell and moving left.
    *   @param  pvStartIndex    Start index of the cell of interest
    *                           in mBOString
    *   @param  pvAttrEndIndex  End index of the cell of interest in
    *                           mBOString
    *   @return     Ending index of data in cell of interest.
    */
    protected int trimRight(int pvStartIndex, int pvAttrStartIndex)
    {
        int lvRightIndex = pvStartIndex - 1;
        while (mPadCharacter == mBOString.charAt(lvRightIndex) &&
                0 < lvRightIndex &&
                pvAttrStartIndex < lvRightIndex)
            lvRightIndex--;
        lvRightIndex++; // Compensates for the last checked character

        return lvRightIndex;
    }

    /**
    *   Append indicated number of instances of the pad character to the left
    *   of the data for the cell of interest.
    *   @param  pvStringBuffer      StringBuffer to which the characters are
    *                               added.  Value changed by this method.
    *   @param  pvPadCharNumber     Number of characters which to add to the
    *                               left of the token data contained in
    *                               pvStringBuffer.
    */
    protected void padLeft(StringBuffer pvStringBuffer, int pvPadCharNumber)
    {
        for (int lvCharIndex = 0; lvCharIndex < pvPadCharNumber; lvCharIndex++)
            pvStringBuffer.insert(0, mPadCharacter);
    }

    /**
    *   Append indicated number of instances of the pad character to the right
    *   of the data for the cell of interest.
    *   @param  pvStringBuffer      StringBuffer to which the characters are
    *                               added.  Value changed by this method.
    *   @param  pvPadCharNumber     Number of characters which to add to the
    *                               right of the token data contained in
    *                               pvStringBuffer.
    */
    protected void padRight(StringBuffer pvStringBuffer, int pvPadCharNumber)
    {
        for (int lvCharIndex = 0; lvCharIndex < pvPadCharNumber; lvCharIndex++)
            pvStringBuffer.append(mPadCharacter);
    }


  /**
   * Parses the app-text of a business object and generates a list of the
   * attributes reserved for child meta objects.
   * @return list of meta object attribute names
   */
  public static List listMOAttrNames( BusinessObjectInterface busObj ) {
      List list = new LinkedList();
      StringTokenizer tok = new StringTokenizer( busObj.getAppText(), ";" );
      while( tok.hasMoreTokens() ) {
          String nextToken = tok.nextToken();
          try {
              int nameValueDelimPosition = nextToken.indexOf( '=' );
              String tag = nextToken.substring( 0, nameValueDelimPosition ).trim().toLowerCase();
              String attributeName = nextToken.substring( nameValueDelimPosition + 1 ).trim();
              if( tag.startsWith( "cw_mo_" ) ) {
                  list.add( attributeName );
              }
          }
          catch( Exception e ) {
            //ignore non-name-value pair elements
          }
      }
      return list;
  }//method listMOAttrNames

}