/* IBM Confidential
 * OCO Source Materials
 * 5724-C11, 5724-D17,5724-E30
 * (C) Copyright IBM Corporation 1997,2002
 * The source code for this program is not published or otherwise
 * divested of its trade secrets, irrespective of what has been
 * deposited with the U.S. Copyright Office.
*/

package com.crossworlds.DataHandlers.text;

import com.crossworlds.DataHandlers.DataHandler;
import com.crossworlds.DataHandlers.Exceptions.CW_BOFormatException;
import com.crossworlds.DataHandlers.Exceptions.CW_EndOfBOException;
import com.crossworlds.DataHandlers.Exceptions.CW_SubscriptionException;

//CW classes
import CxCommon.Exceptions.CxObjectNoSuchAttributeException;
import CxCommon.Exceptions.BusObjSpecNameNotFoundException;
import CxCommon.Exceptions.CxObjectInvalidAttrException;
import CxCommon.BusinessObjectInterface;
import CxCommon.CxObjectContainerInterface;
import CxCommon.CxObjectAttr;
import AppSide_Connector.JavaConnectorUtil;

// java classes
import java.io.InputStream;
import java.io.ByteArrayInputStream;
import java.io.Reader;
import java.util.List;
import java.util.LinkedList;
import java.util.StringTokenizer;



public class namevalue extends DataHandler
{
    private int mDepth;
    private StringBuffer mBOStringBuffer;
    private String mBOString;
    private String[] mCurrTokenPair;

    private final String VerbToken = "Verb";
    private final String BOToken = "BusinessObject";
    private final String AttrCountToken = "AttributeCount";
    private final String ObjectCountToken = "ObjectCount";


    // CR9251 - Removed all of the transient keywords
    protected boolean mValidateAttrCount;
    protected String  mDefaultVerb;
    protected boolean mSkipCxIgnore;
    protected String mCxIgnore;
    protected String mCxBlank;
    protected String mCxBlankValue;

    //for the BOString parsing
    private int mStartIndex;  //of a line
    private int mEndIndex;    //of a line
    private int mLineNum;     //for error messaging
    private int mLineStartIndex;
    private int mLineEndIndex;
    private int mBOStringLength;
    private boolean mEndOfBOString;

    public namevalue()
    {
        mDepth = 0;
    }

    /**
    *   Set the configuration options based on the options set through the
    *   base class.
    */
    private void getConfigOptions()
    {
        traceWrite( "Entering getConfigOptions()", JavaConnectorUtil.LEVEL4);

        // Set the member variables according to the properties set
        //  based on the config object.
        mCxIgnore = getOption("CxIgnore");
        if( mCxIgnore == null || mCxIgnore.length()==0 )
            mCxIgnore = "CxIgnore";
        mCxBlank = getOption("CxBlank");
        if( mCxBlank == null || mCxBlank.length() == 0 )
            mCxBlank = "CxBlank";

        mCxBlankValue = getOption("CxBlankValue");
        if( mCxBlankValue == null || mCxBlankValue.length() == 0 )
            mCxBlankValue = "";

        mDefaultVerb = getOption("DefaultVerb");
        if( mDefaultVerb == null || mDefaultVerb.length() == 0 )
            mDefaultVerb = "Create";

        String lvValidateAttrCount = getOption("ValidateAttrCount");
        if( lvValidateAttrCount.compareTo("true") == 0 )
            mValidateAttrCount = true;
        else if( lvValidateAttrCount.compareTo("false") == 0 )
            mValidateAttrCount = false;
        else
            mValidateAttrCount = true;

        String lvSkipCxIgnore = getOption("SkipCxIgnore");
        if( lvSkipCxIgnore.compareTo("true") == 0 )
            mSkipCxIgnore = true;
        else if( lvSkipCxIgnore.compareTo("false") == 0 )
            mSkipCxIgnore = false;
        else
            mSkipCxIgnore = true;


        traceWrite( "Exiting getConfigOptions", JavaConnectorUtil.LEVEL4);
    }


    /**
    * Resets member variables.  Should be called each time a method in the
    * class is called to do a conversion to or from a business object.
    *
    * @param config Object which can be used to pass in additional
    *                  configuration properties.
    */
    private void clear(Object config)
        throws Exception
    {
        traceWrite( "Entering clear(Object) method", JavaConnectorUtil.LEVEL4);

        if ( config instanceof BusinessObjectInterface )
        {
            traceWrite( "Setting options based on config object attributes",
                        JavaConnectorUtil.LEVEL4);
            setupOptions( (BusinessObjectInterface)config);
        }

        getConfigOptions();

        mStartIndex = 0;
        mEndIndex = 0;
        mLineStartIndex = 0;
        mLineEndIndex = 0;
        mLineNum = 0;
        mBOStringLength = 0;
        mEndOfBOString = false;
        mCurrTokenPair = new String[2];

        traceWrite( "Exiting clear(Object) method", JavaConnectorUtil.LEVEL4);
    }

    /**
    *   Converts the data contained in a Reader object to a CrossWorlds
    *   business object.  Allows configuration options to be set by passing
    *   in a meta object which contains the config information.
    *   @param  serializedData  Reader object which contains the data to be
    *                           converted to a business object.
    *   @param  config          If this parameter is an instance of
    *                           BusinessObjectInterface config options will
    *                           be set based on the default values of the
    *                           attributes.
    */
    public BusinessObjectInterface  getBO(Reader serializedData, Object config)
        throws Exception
    {
        traceWrite( "Entering getBO(Reader, Object)",
                    JavaConnectorUtil.LEVEL4);

        clear(config);

        int conversionCheck;
        char[] temp = new char[2000];
        StringBuffer tempStringBuffer = new StringBuffer(1000);
        String tempString;

        while ( (conversionCheck = serializedData.read (temp)) != -1)
            tempStringBuffer.append(new String (temp, 0, conversionCheck));

        mBOString = tempStringBuffer.toString();
        mBOStringLength = mBOString.length();
        traceWrite( "Converted Reader object to String with length " + mBOStringLength,
                    JavaConnectorUtil.LEVEL5);

        BusinessObjectInterface responseObj;
        responseObj = parseBOString( null, true);

		// <I18N> by Chen Gang (07/22/2002)
		if(responseObj != null && getLocale() != null)
			responseObj.setLocale(getLocale());
		// </I18N>
        traceWrite( "Exiting getBO(Reader, Object)",
                    JavaConnectorUtil.LEVEL4);

        return responseObj;
    }

    /**
    *   Populates a given BusinessObjectInterface with data extracted
    *   from String object. In this case it is up to the caller to create
    *   the BO.  Allows configuration options to be set by passing in a meta
    *   object which contains the config information.
    *   @param  serializedData  String object which contains the data to be
    *                           converted to a business object.
    *   @param  theObj          BusinessObjectInterface to be populated with
    *                           data from Reader object 'serializedData'.
    *   @param  config          If this parameter is an instance of
    *                           BusinessObjectInterface config options will be
    *                           set based on the default values of the
    *                           attributes.
    */
    public void getBO(Reader serializedData, BusinessObjectInterface theObj,
                      Object config)
        throws Exception
    {
       traceWrite( "Entering getBO(Reader, BusinessObjectInterface, Object)",
                   JavaConnectorUtil.LEVEL4);

        clear(config);

        int conversionCheck;
        char[] temp = new char[2000];
        StringBuffer tempStringBuffer = new StringBuffer(1000);
        String tempString;

        while ( (conversionCheck = serializedData.read (temp)) != -1)
            tempStringBuffer.append(new String (temp, 0, conversionCheck));

        mBOString = tempStringBuffer.toString();
        mBOStringLength = mBOString.length();

        traceWrite( "Converted Reader object to String with length " + mBOStringLength,
                    JavaConnectorUtil.LEVEL5);

        //theObj = parseBOString( theObj.getName());
        parseBOString ( theObj, true );

		// <I18N> by Chen Gang (07/22/2002)
		if(theObj != null && getLocale() != null)
			theObj.setLocale(getLocale());
		// </I18N>

        traceWrite( "Exiting getBO(Reader, BusinessObjectInterface, Object)",
                   JavaConnectorUtil.LEVEL4);
    }



    /**
    *   Populates a given BusinessObjectInterface with data extracted
    *   from String object. In this case it is up to the caller to create
    *   the BO.  Allows configuration options to be set by passing in a meta
    *   object which contains the config information.
    *   @param  serializedData  String object which contains the data to be
    *                           converted to a business object.
    *   @param  theObj          BusinessObjectInterface to be populated with
    *                           data from Reader object 'serializedData'.
    *   @param  config          If this parameter is an instance of
    *                           BusinessObjectInterface config options will be
    *                           set based on the default values of the
    *                           attributes.
    */
    public void getBO(String serializedData, BusinessObjectInterface theObj,
            Object config )
        throws Exception
    {
        traceWrite( "Entering getBO(Reader, BusinessObjectInterface, Object)",
                    JavaConnectorUtil.LEVEL4);

        clear(config);
        mBOString = serializedData;
        mBOStringLength = mBOString.length();

        parseBOString( theObj, true );

		// <I18N> by Chen Gang (07/22/2002)
		if(theObj != null && getLocale() != null)
			theObj.setLocale(getLocale());
		// </I18N>

        traceWrite( "Exiting getBO(Reader, BusinessObjectInterface, Object)",
                    JavaConnectorUtil.LEVEL4);
    }

    /**
    *   Creates a String based on the data contained in a BusinessObjectInterface.
    *   Any valid CrossWorlds business object can be used.
    *   Allows configuration options to be set by passing in a meta object
    *   which contains the config information.
    *   @param  theObj      BusinessObjectInterface which contains the data to
    *                       be included in the String.
    *   @param  config      If an instance of BusinessObjectInterface is
    *                       passed in, the default values of the attributes
    *                       are used as options for the formatter.
    *   @return     String which contains the data from theObj in name value
    *               pair format.
    */
   public String getStringFromBO(BusinessObjectInterface theObj,
         Object config)
      throws Exception
   {
      traceWrite( "Entering getStringFromBO(BusinessObjectInterface) for business object "
                    + theObj.getName(),
                    JavaConnectorUtil.LEVEL4);

      // CR 7338: calling 'clear' so that all the properties are initialized
      // properly
      clear (config);

      mDepth = 0;
      mBOStringBuffer = new StringBuffer();

      serializeBO(theObj, true);

      traceWrite( "Exiting getStringFromBO(BusinessObjectInterface) for business object "
                    + theObj.getName(),
                    JavaConnectorUtil.LEVEL4);

      return mBOStringBuffer.toString();
   }

    /**
    *   Creates an InputStream based on the data contained in a BusinessObjectInterface.
    *   Creates a String object first, then converts it to an InputStream to
    *   return to the caller.
    *   Any valid CrossWorlds business object can be used.
    *   Allows configuration options to be set by passing in a meta object
    *   which contains the config information.
    *   @param  theObj      BusinessObjectInterface which contains the data to
    *                       be included in the String.
    *   @param  config      If an instance of BusinessObjectInterface is
    *                       passed in, the default values of the attributes
    *                       are used as options for the formatter.
    *   @return     InputStream which contains the data from theObj in
    *               name value format.
    */
   public InputStream getStreamFromBO(BusinessObjectInterface theObj,
         Object config)
      throws Exception
   {
      traceWrite( "Entering getStreamFromBO(BusinessObjectInterface) for business object "
                    + theObj.getName(),
                    JavaConnectorUtil.LEVEL4);

      String BOstring;

      BOstring = getStringFromBO(theObj, config);

      traceWrite( "Exiting getStreamFromBO(BusinessObjectInterface) for business object "
                    + theObj.getName(),
                    JavaConnectorUtil.LEVEL4);

		// <I18N> by Chen Gang (07/22/2002)
		// return new ByteArrayInputStream( BOstring.getBytes() );
		if(getEncoding() != null) {
			return new ByteArrayInputStream( BOstring.getBytes(getEncoding()));
		} else {
			return new ByteArrayInputStream( BOstring.getBytes());
		}
		// </I18N>
   }

    /**
    *   Build a string with the data contained in the BusinessObjectInterface.
    *   Operates on mBOStringBuffer.
    *   @param  pvBO        Business object to be serialized.
    *   @param  pvIsMainBO  Boolean flag to indicate whether this object is the top
    *                       level business object.  Level of BO used to properly
    *                       format the output (indent correctly).
    */
    protected void serializeBO(BusinessObjectInterface pvBO, boolean pvIsMainBO)
            throws Exception
    {
        traceWrite( "Entering serializeBO(BusinessObjectInterface, boolean) for business object "
                    + pvBO.getName() + " main BO flag " + pvIsMainBO,
                    JavaConnectorUtil.LEVEL4);

        // mDepth is used to determine indention of the name value pairs
        if (pvIsMainBO)
        {
            mDepth = 1;
        }
        else
            ++mDepth;

        try
        {
            traceWrite( "Setting BOName token " + BOToken + " = " +
                        pvBO.getName(), JavaConnectorUtil.LEVEL5);
            this.setSimpleToken(BOToken, pvBO.getName());
            traceWrite( "Setting Verb token " + VerbToken + " = " +
                        pvBO.getVerb(), JavaConnectorUtil.LEVEL5);
            this.setSimpleToken(VerbToken ,pvBO.getVerb());

            // Add token for the attribute count if the option is set for
            //  the Data Handler.
            if( mValidateAttrCount )
                this.setSimpleToken( AttrCountToken,
                            String.valueOf(pvBO.getAttrCount()));

            this.setAttrList(pvBO);
            -- mDepth;
        }
        catch (Exception e)
        {
            throw new Exception("Error serializing BO " + pvBO.getName()
                                + ": " + e.toString());
        }

        traceWrite( "Exiting serializeBO(BusinessObjectInterface, boolean) for business object "
                    + pvBO.getName() + " main BO flag " + pvIsMainBO,
                    JavaConnectorUtil.LEVEL4);

    }

    /**
    *   Adds the name value pairs for the attributes to the serialized business object.
    *   @param pvBO     Business object which contains the data to be serialized.
    */
    protected void setAttrList(BusinessObjectInterface pvBO) throws Exception
    {
        traceWrite( "Entering setAttrList(BusinessObjectInterface) for business object "
                    + pvBO.getName(), JavaConnectorUtil.LEVEL4);

        String lvAttrName = null;
        try
        {
            List moAttrNames = listMOAttrNames( pvBO );

            int lvAttrCount = pvBO.getAttrCount();

            // Loop through the each attribute of the object to add the data to the
            //  serialized representation.
            ATTRIBUTE_WALK: for (int lvAttrIndex = 0; lvAttrIndex < lvAttrCount; ++lvAttrIndex)
            {
                CxObjectAttr lvAttrSpec = pvBO.getAttrDesc(lvAttrIndex);
                lvAttrName = lvAttrSpec.getName();

                // Check if the attribute is a simple attribute or a container
                if (lvAttrSpec.isObjectType())
                {
                    //skip child objects designated as meta objects
                    if( moAttrNames.contains( lvAttrName ) ) {
                      continue ATTRIBUTE_WALK;
                    }

                    if (lvAttrSpec.isMultipleCard())
                    {
                        CxObjectContainerInterface lvAttrMultiCardBOValue =
                            (CxObjectContainerInterface) pvBO.getAttrValue(lvAttrIndex);

                        // If the container is empty, add CxIgnore as the value of the
                        //  attribute.
                        if (lvAttrMultiCardBOValue == null && !mSkipCxIgnore)
                        {
                            traceWrite( "Found empty cardinality n container "
                                        + lvAttrSpec.getName() + " in object " + pvBO.getName(),
                                        JavaConnectorUtil.LEVEL5);

                            this.setSimpleToken(lvAttrSpec.getName(), mCxIgnore);
                        }
                        else if (lvAttrMultiCardBOValue != null)
                        {
                            // Add the attribute name and count to the serialized data
                            int lvObjectCount = lvAttrMultiCardBOValue.getObjectCount();

                            traceWrite( "Found cardinality n container " + lvAttrSpec.getName()
                                        + " with " + lvObjectCount + " contained objects in object "
                                        + pvBO.getName(), JavaConnectorUtil.LEVEL5);

                            this.setSimpleToken(lvAttrSpec.getName(),
                                                new Integer(lvObjectCount));

                            // Loop through the container and add each of the contained
                            //  objects to the serialized data
                            for (int lvContObjectIndex = 0; lvContObjectIndex < lvObjectCount; ++lvContObjectIndex)
                                serializeBO(lvAttrMultiCardBOValue.getBusinessObject(lvContObjectIndex), false);
                        }
                    }
                    else
                    {
                        // Single cardinality container
                        BusinessObjectInterface lvAttrSingleCardBOValue =
                                (BusinessObjectInterface) pvBO.getAttrValue(lvAttrIndex);

                        // If the container is empty, add CxIgnore as the value of the
                        //  attribute.
                        if (lvAttrSingleCardBOValue == null && !mSkipCxIgnore)
                        {
                            traceWrite( "Found empty cardinality 1 container "
                                        + lvAttrSpec.getName() + " in object " + pvBO.getName(),
                                        JavaConnectorUtil.LEVEL5);

                            this.setSimpleToken(lvAttrSpec.getName(), mCxIgnore);
                        }
                        else if (lvAttrSingleCardBOValue != null)
                        {
                            traceWrite( "Found cardinality 1 container " + lvAttrSpec.getName()
                                        + " with contained object in object "
                                        + pvBO.getName(), JavaConnectorUtil.LEVEL5);

                            // Add the name of the attribute and the BO count to the
                            //  serialized object.
                            this.setSimpleToken(lvAttrSpec.getName(), "1");
                            // Add the contained object to the serialized data
                            serializeBO(lvAttrSingleCardBOValue, false);
                        }
                    }
                }
                else
                {
                    // Add the attribute name and value to the serialized data
                    this.setSimpleToken(lvAttrSpec.getName(),
                                        pvBO.getAttrValue(lvAttrIndex));
                }
            }
        }
        catch (CxObjectNoSuchAttributeException e)
        {
            throw new Exception(" Attribute donesn't exist error for BO " + pvBO.getName() + "; "+
                                        e.getMessage());
        }
        catch (Exception e)
        {
            throw new Exception(", Attribute = " + lvAttrName + " has error: " + e.toString());
        }

        traceWrite( "Exiting setAttrList(BusinessObjectInterface) for business object "
                    + pvBO.getName(), JavaConnectorUtil.LEVEL4);
    }

    /**
    *   Adds a single name value pair to mBOStringBuffer based on the token name and
    *   passed in as parameters.
    *   @param  pvToken         This is the name of the token to be added (i.e. right
    *                           side of the equals sign).
    *   @param  pvTokenValue    This is the value of the token to be added (i.e. left
    *                           side of the equals sign).
    */
    protected void setSimpleToken(String pvToken, Object pvTokenValue)
    {
        traceWrite( "Entering setSimpleToken(String, Object) for token " + pvToken,
                    JavaConnectorUtil.LEVEL4);

        if (pvTokenValue != null)  //write as given
        {
            this.indent();
            mBOStringBuffer.append(pvToken);
            mBOStringBuffer.append(" = ");
            if (pvTokenValue.equals(mCxBlankValue))  //CxBlank
                mBOStringBuffer.append(mCxBlank);
            else
                mBOStringBuffer.append(pvTokenValue.toString());
            mBOStringBuffer.append("\n");
        }
        else if (! mSkipCxIgnore)  //pvTokenValue == null
        {
            this.indent();
            mBOStringBuffer.append(pvToken);
            mBOStringBuffer.append(" = ");
            mBOStringBuffer.append(mCxIgnore);
            mBOStringBuffer.append("\n");
        }

        traceWrite( "Exiting setSimpleToken(String, Object) for token " + pvToken,
                    JavaConnectorUtil.LEVEL4);
    }

    /**
    *   Adds as many tabs as the degree of child object depth to
    *   mBOStringBuffer.  Number of tabs determined by member
    *   variable mDepth.
    */
    protected void indent()
    {
        for (int lvTabCount = 0; lvTabCount < mDepth; lvTabCount++)
        {
            mBOStringBuffer.append("\t");
        }
    }

    /**
    *   Parses the serialized business object represenation.  Checks the
    *   BO type against what is contianed in the serialized data if a type is
    *   passed in.  Assumes that the serialized representation has as the
    *   first line the business object name token and the business object verb
    *   token as the second line.
    *   @param  pvBOType    The name of the business object which should be
    *                       created from the current position in the serialized
    *                       data.  Assume it is 'null' only for the top level
    *                       business object, otherwise populated.
    *   @return     BusinessObjectInterface with attributs populated based on
    *               data in the serialized business object.
    */
    protected BusinessObjectInterface parseBOString(BusinessObjectInterface pvBO, boolean isMainBO)
    throws Exception
    {
        traceWrite( "Entering parseBOString(String)",
                    JavaConnectorUtil.LEVEL4);

        //BusinessObjectInterface lvResultBO = null;

        String pvBOType = null;
        if (pvBO != null) {
          pvBOType = pvBO.getName();
        }

        String lvBOName = new String();
        String lvVerb = new String(mCxIgnore);

        try
        {
            // Gets the first token in the serialized data if this is the
            //  first call to parseBOString (top level object).
            // if( pvBOType == null )
            mCurrTokenPair = getNextToken();


            if( mCurrTokenPair[0].compareTo(BOToken)== 0 )
            {
                lvBOName = mCurrTokenPair[1];

                //  Check the given BO name against the name parsed from the
                //      serialized data.
                if (pvBOType != null && ! pvBOType.equals(lvBOName))
                    throw new Exception(" Plain Text Data Handler expects "
                        + pvBOType + " found " + lvBOName + ".\n");

                // First token was the business object name, so we get the
                //  next one.
                mCurrTokenPair = getNextToken();
            }
            // CR 7338:
            // uncommenting this Else part...
            else
            {
                traceWrite( "WARNING: Did not find Business Object name token "
                            + " in serialized business object data",
                            JavaConnectorUtil.LEVEL1);
                if( pvBOType != null )
                    lvBOName = pvBOType;
                else
                    throw new Exception( "No Business Object name token found "
                            + " and no Business Object name sent as a parameter");
            }

            // else
            // CR 7338:
            // remove the above Else
            if (pvBO == null)
              pvBO = JavaConnectorUtil.createBusinessObject(lvBOName);

            // The verb token may not appear for child objects.
            //  There should be an error generated by the caller if the top level object
            //  does not have the verb set.
            if( mCurrTokenPair[0].compareTo(VerbToken) == 0 )
            {
                lvVerb = mCurrTokenPair[1];

                //  Token pair was the verb to use on the business object, so
                //      we get the next token to process.
                mCurrTokenPair = getNextToken();
            }

            //lvResultBO = JavaConnectorUtil.createBusinessObject(lvBOName);

            if ( (!mCxIgnore.equals(lvVerb)) && (!mCxBlank.equals(lvVerb)) )
                pvBO.setVerb(lvVerb);

            // Section to validate the number of attributes written to the file
            //  with the number of attributes in the business object spec.
            if( mValidateAttrCount )
            {
                // Check that the current token is the attribute count
                if( mCurrTokenPair[0].compareTo(AttrCountToken) != 0 )
                {
                    throw new Exception("Missing Attribute Count token at line "
                                + mLineNum);
                }

                traceWrite( "Found attribute count " + mCurrTokenPair[1] +
                            " for business object " + pvBO.getName() +
                            " in serialized data",
                            JavaConnectorUtil.LEVEL5);

                int lvAttrCountInBOString = Integer.parseInt(mCurrTokenPair[1]);
                int lvAttrCount = pvBO.getAttrCount();

                // Check that the attribute count from the serialized data matches the
                //  count in the BO Spec, otherwise generate an error.
                if (lvAttrCount != lvAttrCountInBOString)
                    throw new Exception(" Specified attribute count doesn't match BO definition for BO " + lvBOName);

                mCurrTokenPair = getNextToken();
            }

            //  Parse through the BO attributes in the
            this.parseAttributeList(pvBO);

        }
        catch (BusObjSpecNameNotFoundException e)
        {
            throw new Exception(" Data format error in Plain Text Data Handler:  " + e.toString());
        }
        catch (ArrayIndexOutOfBoundsException e)
        {
            throw new Exception("Name Value Data Handler Internal Error: "
                         + e.getMessage());
        }

        traceWrite( "Exiting parseBOString(String) for business object " + pvBOType,
                    JavaConnectorUtil.LEVEL4);

        return pvBO;
    }

    /**
    *   Parses the list of attributes and data in mBOString to populate the attributes
    *   of the BusinessObjectInterface passed in as a parameter.
    *   Assumes data includes token for ObjectEventId attribute.
    *   @param  pvBO    BusinessObjectInterface which should be populated based on the
    *                   name value pairs in the serialized data.
    */
    protected void parseAttributeList(BusinessObjectInterface pvBO)
        throws Exception
    {
        traceWrite( "Entering parseAttributeList(BusinessObjectInterface) for object "
                    + pvBO.getName(), JavaConnectorUtil.LEVEL4);

        if (mEndOfBOString || pvBO == null)
            throw new Exception(" Plain Text Data Handler found unexpected ending of BO " +
                    pvBO + ".  Please check your data format; ");

        int lvAttrNum = pvBO.getAttrCount();
        boolean hasMoreAttr = true;

        String lvAttrName = null;
        String lvAttrValue = null;
        int lvAttrCount = 0;  //keep track of the number of attributes encountered

        try
        {
            List moAttrNames = listMOAttrNames( pvBO );


            while (hasMoreAttr)
            {
                // Set the attribute name and value locally based on the current token
                //  information.
                lvAttrName = mCurrTokenPair[0];
                lvAttrValue = mCurrTokenPair[1];

                // Get the next token from the serialized data as the current
                //  values are already being processed.
                // mCurrTokenPair = getNextToken();

                int lvAttrIndex = pvBO.getAttributeIndex(lvAttrName);
                CxObjectAttr lvAttrSpec = pvBO.getAttrDesc(lvAttrIndex);

                if (lvAttrSpec.isObjectType())
                {
                    // In this case, the value is actually the number of instances
                    //  of the business object in the child container. This will be
                    //  passed to either getMultipleCard or getSingleCard to process
                    //  the container.
                    String lvBOType = lvAttrSpec.getTypeName();
                    Object lvAttrBOValue = null;

                    if (lvAttrSpec.isMultipleCard())
                        this.getMultipleCard(pvBO, lvAttrIndex, lvBOType, lvAttrValue);
                    else
                        this.getSingleCard(pvBO, lvAttrIndex, lvBOType, lvAttrValue);

                    //if child object is a meta object, clear it
                    if( moAttrNames.contains( lvAttrName ) ) {
                      pvBO.setAttrValue( lvAttrIndex, JavaConnectorUtil.getIgnoreValue() );
                    }

                    ++lvAttrCount;
                }
                else
                {
                    //simple attribute
                    if (! lvAttrValue.equals(mCxIgnore))
                        pvBO.setAttrValue(lvAttrIndex, lvAttrValue);

                    if (lvAttrSpec.isObjectEventIdAttr())
                        hasMoreAttr = false;   //this is the only normal exit
                    else
                        ++lvAttrCount;

                    mCurrTokenPair = getNextToken();
                }

                if (mEndOfBOString && hasMoreAttr)
                    throw new Exception(" Unexpected ending of BO " + pvBO.getName() + ".  Please check your data format; ");
                if (hasMoreAttr && lvAttrCount >= lvAttrNum)
                    throw new Exception(" Possible redundent BO attributes; BO " + pvBO.getName());

            }
        }
        catch (CxObjectNoSuchAttributeException e)
        {
            //found format error of this BO
            throw new Exception(" Attribute doesn't exist error: " + e.getMessage() + ". ");
        }
        catch (CxObjectInvalidAttrException e)
        {
            throw new Exception(" Attempt to access invalid attribute: " +  e.getMessage());
        }

    }

    /**
    *   Populates a single cardinatlity child business object with the token
    *   values found in the delimited string.
    *   @param  pvParentBO              Parent business object.
    *   @param  pvParentAttrIndex       Attribute index of the child to be
    *                                   created.
    *   @param  pvBOType                Type of the child business object.
    *   @param  pvObjectCountString     Number of instances of the child
    *                                   object as parsed from the delimited
    *                                   string.
    */
    protected void  getSingleCard(BusinessObjectInterface pvParentBO,
                                  int pvParentAttrIndex, String pvBOType,
                                  String pvObjectCountString)
        throws Exception
    {
        traceWrite( "Entering getSingleCard for parentBO "
                    + pvParentBO.getName() + " index " + pvParentAttrIndex
                    + " with count " + pvObjectCountString,
                    JavaConnectorUtil.LEVEL4);

        try
        {
            BusinessObjectInterface lvChildBO = null;

            // Check the object count token
            // If it does not match "1", assume that the child object should
            //  be null
            if (pvObjectCountString.equals("1"))
            {
                // The delimited string contains a single instance of the
                //  child object
                traceWrite( "Found single instance of child object "
                            + pvBOType + " in delimited string",
                            JavaConnectorUtil.LEVEL5);

                lvChildBO = JavaConnectorUtil.createBusinessObject(pvBOType);
                //lvChildBO = parseBOString(pvBOType);
                parseBOString (lvChildBO, false);

                pvParentBO.setAttrValue(pvParentAttrIndex, lvChildBO);
            }
            else if ( pvObjectCountString.equals(mCxIgnore) ||
                        pvObjectCountString.equals("0"))
            {
                // This case is included to validate the object count token is valid
                traceWrite( "Encountered null child object " + pvBOType
                            + " for parent " + pvParentBO.getName(),
                            JavaConnectorUtil.LEVEL5);

                mCurrTokenPair = getNextToken();
            }
            else
                throw new CW_BOFormatException(" Line: "
                        + mLineNum +
                        " BO object count error.  This count should be 1, 0, or CxIgnore. " );

        }
        catch (CxObjectInvalidAttrException e)
        {
            throw new Exception(" Column: " + mLineNum +
                                ": try to access to the wrong attribute. "
                                +  e.getMessage() + ". ");
        }
        catch (CxObjectNoSuchAttributeException e)
        {
            throw new Exception(" Column: " + mLineNum +
                                ": try to access to the wrong attribute. " +  e.getMessage() + ". ");
        }
        catch (BusObjSpecNameNotFoundException e)
        {
            throw new Exception("Invalid BusObjSpec name found: "
                                + e.getMessage());
        }

        traceWrite( "Exiting getSingleCard for parentBO " + pvParentBO.getName() + " index " + pvParentAttrIndex
                    + " with count " + pvObjectCountString,
                    JavaConnectorUtil.LEVEL4);
    }

    /**
    *   Populates a child container based on token values in the delimited
    *   string.
    *   Expects all objects in the child container to be of the same type.
    *   @param  pvParentBO              Parent business object.
    *   @param  pvParentAttrIndex       Attribute index of the child to be
    *                                   created.
    *   @param  pvBOType                Type of the child business object.
    *   @param  pvObjectCountString     Number of instances of the child
    *                                   object as parsed from the delimited
    *                                   string.
    */
    protected void getMultipleCard(BusinessObjectInterface pvParentBO,
                                   int pvParentAttrIndex, String pvBOType,
                                   String pvObjectCountString)
        throws Exception
    {
        traceWrite( "Entering getMultipleCard for parentBO "
                    + pvParentBO.getName() + " index " + pvParentAttrIndex
                    + " with count " + pvObjectCountString,
                    JavaConnectorUtil.LEVEL4);

        try
        {
            if ( pvObjectCountString.equals(mCxIgnore) )
            {
                traceWrite( "Count string for container " + pvBOType
                            + " in parent " + pvParentBO.getName()
                            + " is CxIgnore value",
                            JavaConnectorUtil.LEVEL5);

                mCurrTokenPair = getNextToken();
            }
            else
            {
                int lvObjectCount = Integer.parseInt(pvObjectCountString);

                if ( lvObjectCount == 0)
                {
                    traceWrite( "Count string for container " + pvBOType
                                + " in parent " + pvParentBO.getName()
                                + " is 0",
                                JavaConnectorUtil.LEVEL5);
                  //CR 16215
                    mCurrTokenPair = getNextToken();
                  //end CR 16215
                }
                else if (lvObjectCount > 0)
                {
                    // There is at least one instance of the object in the
                    //  delimited string

                    traceWrite( "Found " + lvObjectCount
                                + " instances of child object pvBOType "
                                + " in parent object " + pvParentBO.getName(),
                                JavaConnectorUtil.LEVEL5);



                    // For each instance of the child object, parse the
                    //  attribute list, and then add the object container
                    //  to the parent.

                    // CR 7889
                    CxObjectContainerInterface container =
                                               JavaConnectorUtil.createContainer(pvBOType);
                    for (int lvObjectIndex = 0; lvObjectIndex < lvObjectCount; lvObjectIndex++)
                    {
                        BusinessObjectInterface lvChildBO =
                                                JavaConnectorUtil.createBusinessObject (pvBOType);
                        parseBOString(lvChildBO, false);
                        container.insertBusinessObject(lvChildBO);
                    }
                    pvParentBO.setAttrValue(pvParentAttrIndex, container);
                    //end CR 7889
                }
            }
        }
        catch (CxObjectNoSuchAttributeException e)
        {
            throw new Exception(" Line: " + mLineNum +
                                        ": formatting error; " + e.getMessage() + ". ");
        }
        catch (CxObjectInvalidAttrException e)
        {
            throw new Exception(" Column: " + mLineNum +
                ": formatting error: Attribute miss alignment; " + e.getMessage() + ". ");
        }
        catch (NumberFormatException e)
        {
            throw new Exception(" Column: " + mLineNum +
                ": formatting error:  cannot recognize number format; " + e.getMessage() + ". ");
        }
        catch (BusObjSpecNameNotFoundException e)
        {
            throw new Exception("Invalid BusObjSpec name found: " + e.getMessage());
        }

        traceWrite( "Exiting getMultipleCard for parentBO " + pvParentBO.getName() + " index " + pvParentAttrIndex
                    + " with count " + pvObjectCountString,
                    JavaConnectorUtil.LEVEL4);

    }

    /**
    *   Method to read the name and value of the next token in the serialized
    *   name value data.
    *   @returns        An array of strings with exactly two elements, the first
    *                   being the name of the pair, and the second the value.
    */
    private String[] getNextToken() throws Exception
    {
        String[] tokenPair = new String[2];

        // This moves the current location in the serialized BO string to the
        //  next line.
        try
        {
            getNextLine();

            tokenPair[0] = getLeftHandValue();
            tokenPair[1] = getRightHandValue();
        }
        catch( CW_EndOfBOException e)
        {
            traceWrite( "Found end of the serialized data",
                        JavaConnectorUtil.LEVEL5);
            tokenPair[0] = null;
            tokenPair[1] = null;
        }

        return tokenPair;
    }



  /** local method: start and end delimiting index of line
   *  Expect mBOString in form of delimited by '\n' and empty
   *  line is not skipped.  But line number is still properly
   *  counted.
   */
  protected void getNextLine() throws CW_EndOfBOException
  {
    boolean lvLineEmpty = true;

    if (mLineEndIndex >= mBOStringLength || mEndOfBOString)
      throw new CW_EndOfBOException("End of serialized data.");

    while (lvLineEmpty )
    {
      mLineStartIndex = mLineEndIndex;
      mLineEndIndex = mBOString.indexOf('\n', mLineStartIndex);
      ++mLineNum;

      if (mLineEndIndex > mLineStartIndex)  //at the end, the two willb be equal
      {
        if (! this.emptyLine(mLineStartIndex, mLineEndIndex))
          lvLineEmpty = false;
        else if (mLineEndIndex >= mBOStringLength)
          mEndOfBOString = true;
      }
      mLineEndIndex++; //skip over the '\n' character
    }

    mStartIndex = mLineStartIndex;
    mEndIndex = mLineEndIndex;
    //the String.trim() is cann't get over some carriage return characters
    //trim left
    while (Character.isWhitespace(mBOString.charAt(mStartIndex)) &&
           mStartIndex < mLineEndIndex)
      mStartIndex++ ;

    //trim right
    while (Character.isWhitespace(mBOString.charAt(mEndIndex - 1)) &&
           mStartIndex < mEndIndex)
      mEndIndex--;


  }

  /** operates on mBOString
   */
  protected boolean emptyLine(int pvLeftIndex, int pvRightIndex)
  {
    boolean lvResult = true;

    int lvLIndex = pvLeftIndex;
    int lvRIndex = pvRightIndex;

    while (Character.isWhitespace(mBOString.charAt(lvLIndex)) &&
       lvLIndex < mLineEndIndex)
      lvLIndex++;

    while (Character.isWhitespace(mBOString.charAt(lvRIndex)) &&
           mStartIndex < lvRIndex)
      lvRIndex--;

    if (lvRIndex > lvLIndex)
      lvResult = false;

    return lvResult;
  }
  /** local method: get string fragment of a line in front of
   *  equal sign
   *  return trimed string fragment
   */
  protected String getLeftHandValue()
  {
    String lvResult = null;

    int lvEqualSignIndex = mBOString.indexOf('=', mStartIndex);
    lvResult = mBOString.substring(mStartIndex, lvEqualSignIndex);
    lvResult = lvResult.trim();
    return lvResult;
  }

  /** local method: get string fragment of a line after the
   *  equal sign
   *  return trimed string fragment
   */
  protected String getRightHandValue()
  {
    String lvResult = null;

    int lvEqualSignIndex = mBOString.indexOf('=', mStartIndex);
    lvResult = mBOString.substring(lvEqualSignIndex + 1, mEndIndex);
    lvResult = lvResult.trim();
    //trime right
    return lvResult;
  }


  /**
   * Parses the app-text of a business object and generates a list of the
   * attributes reserved for child meta objects.
   * @return list of meta object attribute names
   */
  public static List listMOAttrNames( BusinessObjectInterface busObj ) {
      List list = new LinkedList();
      StringTokenizer tok = new StringTokenizer( busObj.getAppText(), ";" );
      while( tok.hasMoreTokens() ) {
          String nextToken = tok.nextToken();
          try {
              int nameValueDelimPosition = nextToken.indexOf( '=' );
              String tag = nextToken.substring( 0, nameValueDelimPosition ).trim().toLowerCase();
              String attributeName = nextToken.substring( nameValueDelimPosition + 1 ).trim();
              if( tag.startsWith( "cw_mo_" ) ) {
                  list.add( attributeName );
              }
          }
          catch( Exception e ) {
            //ignore non-name-value pair elements
          }
      }
      return list;
  }//method listMOAttrNames

}
